from dataclasses import fields

from nicegui import ui
from session import backend_session, endpoint

import ieee_2030_5.models as m

columns = [
    {'name': 'property', 'label': 'Property', 'field': 'property', 'required': False},
    {'value': 'value', 'label': 'Value', 'field': 'property', 'required': False},
]

def show_end_device(enddevice: m.EndDevice):
    ui.label(enddevice.href)

def show_list(enddevices: m.EndDeviceList):
    if len(enddevices.EndDevice) == 0:
        ui.label(f"No EndDevice in list. all={enddevices.all}, results={enddevices.results}")
    else:
        
        for ed in enddevices.EndDevice:
            for fld in fields(ed):
                with ui.row():
                    value = getattr(ed, fld.name)
                    if value and fld.type in ('Optional[str]', 'str'):
                        if isinstance(value, str):
                            ui.label(f"{fld.name} -> {value}")    
                        else:
                            ui.label(f"{fld.name} -> {value.decode('utf-8')}")
                    elif value and fld.type in ('int'):
                        ui.label(f"{fld.name} -> {value}")
                    elif value:
                        
                        if isinstance(value, bytes):
                            ui.label(f"{fld.name} -> {value}")
                        elif isinstance(value, int):
                            ui.label(f"{fld.name} -> {value}")
                        elif getattr(value, 'href'):
                            ui.label(f"{fld.name} -> {getattr(value, 'href')}")
                        else:
                            ui.label(f"Type is: {fld.type}")
                    # value = getattr(ed, fld.name)
                    # if value and value.decode('utf-8'):
                    #     ui.label(f"{fld.name} => {value.decode('utf-8')}")

       
def add_end_device():
    
    def submit_end_device(id, pin):
        ui.label(f"Posted {id} and {pin}")
    
    with ui.column():
        
        id = ui.input(label="Id")
        pin = ui.input(label="Pin")
        ui.button("Save", on_click=lambda a: submit_end_device(id.value, pin.value))


from __future__ import annotations

import atexit
import http
import logging
import ssl
import threading
import xml.dom.minidom
from http.client import HTTPSConnection
from os import PathLike
from pathlib import Path
from threading import Timer
from typing import Dict, Optional, Tuple, Any

import werkzeug.middleware.lint
import xsdata

import ieee_2030_5.models as m
import ieee_2030_5.utils as utils
import ieee_2030_5.utils.tls_wrapper as tls

_log = logging.getLogger(__name__)
_log_req_resp = logging.getLogger(__name__ + ".request")


class IEEE2030_5_Client:
    clients: set[IEEE2030_5_Client] = set()

    # noinspection PyUnresolvedReferences
    def __init__(self,
                 cafile: PathLike,
                 server_hostname: str,
                 keyfile: PathLike,
                 certfile: PathLike,
                 server_ssl_port: Optional[int] = 443,
                 debug: bool = True):

        cafile = cafile if isinstance(cafile, PathLike) else Path(cafile)
        keyfile = keyfile if isinstance(keyfile, PathLike) else Path(keyfile)
        certfile = certfile if isinstance(certfile, PathLike) else Path(certfile)

        self._key = keyfile
        self._cert = certfile
        self._ca = cafile

        assert cafile.exists(), f"cafile doesn't exist ({cafile})"
        assert keyfile.exists(), f"keyfile doesn't exist ({keyfile})"
        assert certfile.exists(), f"certfile doesn't exist ({certfile})"

        self._ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        self._ssl_context.check_hostname = False
        self._ssl_context.verify_mode = ssl.CERT_OPTIONAL    #  ssl.CERT_REQUIRED
        self._ssl_context.load_verify_locations(cafile=cafile)

        # Loads client information from the passed cert and key files. For
        # client side validation.
        self._ssl_context.load_cert_chain(certfile=certfile, keyfile=keyfile)

        self._http_conn = HTTPSConnection(host=server_hostname,
                                          port=server_ssl_port,
                                          context=self._ssl_context)
        self._device_cap: Optional[m.DeviceCapability] = None
        self._mup: Optional[m.MirrorUsagePointList] = None
        self._upt: Optional[m.UsagePointList] = None
        self._edev: Optional[m.EndDeviceListLink] = None
        self._end_devices: Optional[m.EndDeviceListLink] = None
        self._fsa_list: Optional[m.FunctionSetAssignmentsListLink] = None
        self._debug = debug
        self._dcap_poll_rate: int = 0
        self._dcap_timer: Optional[Timer] = None
        self._disconnect: bool = False
        self._tls = tls.OpensslWrapper

        IEEE2030_5_Client.clients.add(self)

    @property
    def http_conn(self) -> HTTPSConnection:
        if self._http_conn.sock is None:
            self._http_conn.connect()
        return self._http_conn

    def register_end_device(self) -> str:
        lfid = utils.get_lfdi_from_cert(self._cert)
        sfid = utils.get_sfdi_from_lfdi(lfid)
        response = self.__post__(dcap.EndDeviceListLink.href,
                                 data=utils.dataclass_to_xml(m.EndDevice(sFDI=sfid)))
        print(response)

        if response.status in (200, 201):
            return response.headers.get("Location")

        raise werkzeug.exceptions.Forbidden()

    def get(self, href):
        return self.__get_request__(href)

    def is_end_device_registered(self, end_device: m.EndDevice, pin: int) -> bool:
        reg = self.registration(end_device)
        return reg.pIN == pin

    def new_uuid(self, url: str = "/uuid") -> str:
        res = self.__get_request__(url)
        return res

    def end_devices(self) -> m.EndDeviceList:
        if not self._device_cap:
            self.device_capability()

        self._end_devices = self.__get_request__(self._device_cap.EndDeviceListLink.href)
        return self._end_devices

    def end_device(self, index: Optional[int] = 0) -> m.EndDevice:
        if not self._end_devices:
            self.end_devices()

        return self._end_devices.EndDevice[index]

    def self_device(self) -> m.EndDevice:
        if not self._device_cap:
            self.device_capability()

        return self.__get_request__(self._device_cap.SelfDeviceLink.href)

    def function_set_assignment_list(self,
                                     edev_index: Optional[int] = 0
                                     ) -> m.FunctionSetAssignmentsList:
        fsa_list = self.__get_request__(
            self.end_device(edev_index).FunctionSetAssignmentsListLink.href)
        return fsa_list

    def function_set_assignment(self,
                                edev_index: Optional[int] = 0,
                                fsa_index: Optional[int] = 0) -> m.FunctionSetAssignments:
        fsa_list = self.function_set_assignment_list(edev_index)
        return fsa_list.FunctionSetAssignments[fsa_index]

    def der_list(self, edev_index: Optional[int] = 0) -> m.DERList:
        der_list = self.__get_request__(self.end_device(edev_index).DERListLink.href)
        return der_list

    def poll_timer(self, fn, args):
        if not self._disconnect:
            _log.debug(threading.currentThread().name)
            fn(args)
            threading.currentThread().join()

    def device_capability(self, url: str = "/dcap") -> m.DeviceCapability:
        self._device_cap: m.DeviceCapability = self.__get_request__(url)
        if self._device_cap.pollRate is not None:
            self._dcap_poll_rate = self._device_cap.pollRate
        else:
            self._dcap_poll_rate = 600

        _log.debug(f"devcap id {id(self._device_cap)}")
        _log.debug(threading.currentThread().name)
        _log.debug(f"DCAP: Poll rate: {self._dcap_poll_rate}")
        # self._dcap_timer = Timer(self._dcap_poll_rate, self.poll_timer, (self.device_capability, url))
        # self._dcap_timer.start()
        return self._device_cap

    def time(self) -> m.Time:
        timexml = self.__get_request__(self._device_cap.TimeLink.href)
        return timexml

    def der_program_list(self,
                         edev_index: Optional[int] = 0,
                         fsa_index: Optional[int] = 0) -> m.DERProgramList:
        fsa = self.function_set_assignment(edev_index, fsa_index)
        derp_list = self.__get_request__(fsa.DERProgramListLink.href)
        return derp_list

    def der_program(self,
                    edev_index: Optional[int] = 0,
                    fsa_index: Optional[int] = 0,
                    derp_index: Optional[int] = 0) -> m.DERProgram:
        derp_list = self.der_program_list(edev_index, fsa_index)
        return derp_list.DERProgram[derp_index]

    def mirror_usage_point_list(self) -> m.MirrorUsagePointList:
        self._mup = self.__get_request__(self._device_cap.MirrorUsagePointListLink.href)
        return self._mup

    def usage_point_list(self) -> m.UsagePointList:
        self._upt = self.__get_request__(self._device_cap.UsagePointListLink.href)
        return self._upt

    def registration(self, end_device: m.EndDevice) -> m.Registration:
        reg = self.__get_request__(end_device.RegistrationLink.href)
        return reg

    def timelink(self):
        if self._device_cap is None:
            raise ValueError("Request device capability first")
        return self.__get_request__(url=self._device_cap.TimeLink.href)

    def disconnect(self):
        self._disconnect = True
        if self._dcap_timer:
            self._dcap_timer.cancel()
        IEEE2030_5_Client.clients.remove(self)

    def request(self, endpoint: str, body: dict = None, method: str = "GET", headers: dict = None):

        if method.upper() == 'GET':
            return self.__get_request__(endpoint, body, headers=headers)

        if method.upper() == 'POST':
            print("Doing post")
            return self.__post__(endpoint, body, headers=headers)

    def create_mirror_usage_point(self, mirror_usage_point: m.MirrorUsagePoint) -> Tuple[int, str]:
        data = utils.dataclass_to_xml(mirror_usage_point)
        resp = self.__post__(self._device_cap.MirrorUsagePointListLink.href, data=data)
        return resp.status, resp.headers['Location']

    def create_mirror_meter_reading(self, mirror_usage_point_href: str,
                                    mirror_meter_reading: m.MirrorMeterReading) -> Tuple[int, str]:
        data = utils.dataclass_to_xml(mirror_meter_reading)
        resp = self.__post__(mirror_usage_point_href, data=data)
        return resp.status, resp.headers['Location']

    def post(self, url: str, data: Any, headers: Optional[Dict[str, str]] = None):
        response = self.__post__(url, data, headers=headers)

    def __post__(self, url: str, data=None, headers: Optional[Dict[str, str]] = None):
        if not headers:
            headers = {'Content-Type': 'text/xml'}

        self.http_conn.request(method="POST", headers=headers, url=url, body=data)
        response = self._http_conn.getresponse()
        # response_data = response.read().decode("utf-8")

        return response

    def __get_request__(self, url: str, body=None, headers: dict = None):
        if headers is None:
            headers = {"Connection": "keep-alive", "keep-alive": "timeout=30, max=1000"}

        if self._debug:
            print(f"----> GET REQUEST")
            print(f"url: {url} body: {body}")
        self.http_conn.request(method="GET", url=url, body=body, headers=headers)
        response = self._http_conn.getresponse()
        response_data = response.read().decode("utf-8")
        print(response.headers)

        response_obj = None
        try:
            response_obj = utils.xml_to_dataclass(response_data)
            resp_xml = xml.dom.minidom.parseString(response_data)
            if resp_xml and self._debug:
                print(f"<---- GET RESPONSE")
                print(f"{response_data}")    # toprettyxml()}")

        except xsdata.exceptions.ParserError as ex:
            if self._debug:
                print(f"<---- GET RESPONSE")
                print(f"{response_data}")
            response_obj = response_data

        return response_obj

    def __close__(self):
        self._http_conn.close()
        self._ssl_context = None
        self._http_conn = None

    def put(self, url: str, data: Any, headers: Optional[Dict[str, str]] = None):
        response = self.__put__(url, data, headers=headers)

    def __put__(self, url: str, data: Any, headers: Optional[Dict[str, str]] = None):
        if not headers:
            headers = {'Content-Type': 'text/xml'}

        if self._debug:
            _log_req_resp.debug(f"----> PUT REQUEST\nurl: {url}\nbody: {data}")

        try:
            self.http_conn.request(method="PUT", headers=headers, url=url, body=data)
        except http.client.CannotSendRequest as ex:
            self.http_conn.close()
            _log.debug("Reconnecting to server")
            self.http_conn.request(method="PUT", headers=headers, url=url, body=data)

        response = self._http_conn.getresponse()
        return response

    def __post__(self, url: str, data=None, headers: Optional[Dict[str, str]] = None):
        if not headers:
            headers = {'Content-Type': 'text/xml'}

        if self._debug:
            _log_req_resp.debug(f"----> POST REQUEST\nurl: {url}\nbody: {data}")

        self.http_conn.request(method="POST", headers=headers, url=url, body=data)
        response = self._http_conn.getresponse()
        response_data = response.read().decode("utf-8")
        # response_data = response.read().decode("utf-8")
        if response_data and self._debug:
            _log_req_resp.debug(f"<---- POST RESPONSE\n{response_data}")

        return response


# noinspection PyTypeChecker
def __release_clients__():
    for x in IEEE2030_5_Client.clients:
        x.__close__()
    IEEE2030_5_Client.clients = None


atexit.register(__release_clients__)

#
# ssl_context = ssl.create_default_context(cafile=str(SERVER_CA_CERT))
#
#
# con = HTTPSConnection("me.com", 8000,
#                       key_file=str(KEY_FILE),
#                       cert_file=str(CERT_FILE),
#                       context=ssl_context)
# con.request("GET", "/dcap")
# print(con.getresponse().read())
# con.close()

if __name__ == '__main__':
    SERVER_CA_CERT = Path("~/tls/certs/ca.pem").expanduser().resolve()
    KEY_FILE = Path("~/tls/private/dev1.pem").expanduser().resolve()
    CERT_FILE = Path("~/tls/certs/dev1.pem").expanduser().resolve()

    headers = {'Connection': 'Keep-Alive', 'Keep-Alive': "max=1000,timeout=30"}

    h = IEEE2030_5_Client(cafile=SERVER_CA_CERT,
                          server_hostname="127.0.0.1",
                          server_ssl_port=8443,
                          keyfile=KEY_FILE,
                          certfile=CERT_FILE,
                          debug=True)
    # h2 = IEEE2030_5_Client(cafile=SERVER_CA_CERT, server_hostname="me.com", ssl_port=8000,
    #                        keyfile=KEY_FILE, certfile=KEY_FILE)
    dcap = h.device_capability()
    end_devices = h.end_devices()

    if not end_devices.all > 0:
        print("registering end device.")
        ed_href = h.register_end_device()
    my_ed = h.end_devices()
    my_fsa = h.function_set_assignment()
    my_program = h.der_program()

    # ed = h.end_devices()[0]
    # resp = h.request("/dcap", headers=headers)
    # print(resp)
    # resp = h.request("/dcap", headers=headers)
    # print(resp)
    #dcap = h.device_capability()
    # get device list
    #dev_list = h.request(dcap.EndDeviceListLink.href).EndDevice

    #ed = h.request(dev_list[0].href)
    #print(ed)
    #
    # print(dcap.mirror_usage_point_list_link)
    # # print(h.request(dcap.mirror_usage_point_list_link.href))
    # print(h.request("/dcap", method="post"))

    # tl = h.timelink()
    #print(IEEE2030_5_Client.clients)



# from __future__ import annotations

# import io
# import logging
# import os
# import sys

# from dotenv import load_dotenv
# from fastapi import FastAPI
# from nicegui import ui

# from .pages import Pages, show_global_header

# logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
# _log = logging.getLogger(__name__)

# app = FastAPI()

# __all__ = [
#     "run_gui"
# ]

# def initialize_app():

#     show_global_header(Pages.HOME)

#     ui.markdown("""

# # Welcome to the IEEE 2030.5 Admin GUI

# This is a work in progress""")

#     # device_list = get_end_devices()
#     # certs = get_certs()

#     # print(certs)
#     # print(device_list)
#     # row_data = []
#     # for ed in device_list.EndDevice:
#     #     row = dict(name=f'{ed.lFDI}',
#     #             cert=f'<a href="/admin/cert/{ed.lFDI}">cert</a>')
#     #     row_data.append(row)


#     # ui.aggrid({
#     #     'columnDefs': [
#     #         {'headerName': '', 'field': 'cert'},
#     #         {'headerName': 'Name', 'field': 'name'},
#     #     ],
#     #     'rowData': row_data,
#     # }, html_columns=[0])
# def run_gui():
#     out_file = open(sys.stdout.fileno(), 'wb', 0)
#     sys.stdout = io.TextIOWrapper(out_file, write_through=True)
#     from pathlib import Path

#     PREFIX = "2030_5"

#     # Allow custmization of the 2030_5_ENV_PATH
#     env_path = Path(os.environ.get(f"{PREFIX}_ENV_PATH",
#                                    str(Path(__file__).parent / ".env")))


#     load_dotenv(dotenv_path=env_path.as_posix())

#     required_env = [
#         f"{PREFIX}_HOST",
#         f"{PREFIX}_PORT",
#         f"{PREFIX}_CLIENT_KEY",
#         f"{PREFIX}_CLIENT_CERT",
#         f"{PREFIX}_CA_CERT"
#     ]

#     for x in required_env:
#         if os.getenv(x) is None or os.getenv(x).strip() == "":
#             sys.stderr.write(f'Missing {x} in .env file\n')
#             sys.exit(1)

#     logging.basicConfig(filename="gui-server.log", level=logging.DEBUG)

#     initialize_app()

#     ui.run(show=False)
#     # from pathlib import Path

#     # env_path = Path(__file__).parent / ".env"

#     # load_dotenv(dotenv_path=env_path.as_posix())

#     # PREFIX = "2030_5"
#     # required_env = [
#     #     f"{PREFIX}_HOST",
#     #     f"{PREFIX}_PORT",
#     #     f"{PREFIX}_CLIENT_KEY",
#     #     f"{PREFIX}_CLIENT_CERT",
#     #     f"{PREFIX}_CA_CERT"
#     # ]

#     # for x in required_env:
#     #     if os.getenv(x) is None or os.getenv(x).strip() == "":
#     #         sys.stderr.write(f'Missing {x} in .env file\n')
#     #         sys.exit(1)

#     # logging.basicConfig(level=logging.DEBUG)

#     # initialize_app()


#     # ui.run(show=False)





# # ui.html("""
# #     <table>
# #         <tr>
# #             <td>1</td>
# #             <td>2</td>
# #         </tr>
# #     </table>
# # """)


# # columns = [
# #     {'name': 'name', 'label': 'Name', 'field': 'name', 'required': True},
# #     {'name': 'age', 'label': 'Age', 'field': 'age', 'sortable': True},
# # ]
# # rows = [
# #     {'id': 0, 'name': 'Alice', 'age': 18},
# #     {'id': 1, 'name': 'Bob', 'age': 21},
# #     {'id': 2, 'name': 'Lionel', 'age': 19},
# #     {'id': 3, 'name': 'Michael', 'age': 32},
# #     {'id': 4, 'name': 'Julie', 'age': 12},
# #     {'id': 5, 'name': 'Livia', 'age': 25},
# #     {'id': 6, 'name': 'Carol'},
# # ]

# # with ui.table(title='My Team', columns=columns, rows=rows, selection='multiple', pagination=10).classes('w-96') as table:
# #     with table.add_slot('top-right'):
# #         with ui.input(placeholder='Search').props('type=search').bind_value(table, 'filter').add_slot('append'):
# #             ui.icon('search')
# #     with table.add_slot('body'):
# #         with table.row() as row:
# #             with table.cell() as cell:
# #                 pprint(row.props)
# #                 ui.link('Download', f'https://nicegui.io?{row.props.id}')
# #             with table.cell():
# #                 ui.link('Download', 'https://nicegui.io')
# #             with table.cell():
# #                 ui.link('Download', 'https://nicegui.io')
# #     with table.add_slot('bottom-row'):
# #         with table.row():
# #             with table.cell():
# #                 ui.button(on_click=lambda: (
# #                     table.add_rows({'id': time.time(), 'name': new_name.value, 'age': new_age.value}),
# #                     new_name.set_value(None),
# #                     new_age.set_value(None),
# #                 ), icon='add').props('flat fab-mini')
# #             with table.cell():
# #                 new_name = ui.input('Name')
# #             with table.cell():
# #                 new_age = ui.number('Age')

# # ui.label().bind_text_from(table, 'selected', lambda val: f'Current selection: {val}')
# # ui.button('Remove', on_click=lambda: table.remove_rows(*table.selected)) \
# #     .bind_visibility_from(table, 'selected', backward=lambda val: bool(val))

#     #ui.run(show=False)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from typing import List, NamedTuple, Optional, Union
import ieee_2030_5.models as m

EDEV = "edev"
DCAP = "dcap"
UTP = "upt"
MUP = "mup"
DRP = "drp"
SDEV = "sdev"
MSG = "msg"
DER = "der"
CURVE = "dc"
RSPS = "rsps"
LOG = "log"
DERC = "derc"
DDERC = "dderc"
DERCA = "derca"
FSA = "fsa"
TIME = "tm"

DER_PROGRAM = "derp"
# DER Available
DER_AVAILABILITY = "dera"
# DER Status
DER_STATUS = "ders"
DER_CONTROL_ACTIVE = DERCA
DER_CAPABILITY = "dercap"
# Settings
DER_SETTINGS = "derg"
END_DEVICE_REGISTRATION = "rg"
END_DEVICE_STATUS = "dstat"
END_DEVICE_FSA = FSA
END_DEVICE_POWER_STATUS = "ps"
END_DEVICE_LOG_EVENT_LIST = "lel"
END_DEVICE_INFORMATION = "di"

DEFAULT_TIME_ROOT = f"/{TIME}"
DEFAULT_DCAP_ROOT = f"/{DCAP}"
DEFAULT_EDEV_ROOT = f"/{EDEV}"
DEFAULT_UPT_ROOT = f"/{UTP}"
DEFAULT_MUP_ROOT = f"/{MUP}"
DEFAULT_DRP_ROOT = f"/{DRP}"
DEFAULT_SELF_ROOT = f"/{SDEV}"
DEFAULT_MESSAGE_ROOT = f"/{MSG}"
DEFAULT_DER_ROOT = f"/{DER}"
DEFAULT_CURVE_ROOT = f"/{CURVE}"
DEFAULT_RSPS_ROOT = f"/{RSPS}"
DEFAULT_LOG_EVENT_ROOT = f"/{LOG}"
DEFAULT_FSA_ROOT = f"/{FSA}"
DEFAULT_DERP_ROOT = f"/{DER_PROGRAM}"

SEP = "_"
MATCH_REG = "[a-zA-Z0-9_]*"

# Used as a sentinal value when we only want the href of the root
NO_INDEX = -1


class HrefParser:

    def __init__(self, href: str):
        self.href = href
        self._split = href.split(SEP)

    def has_index(self) -> bool:
        """This function returns true if there is an index on the primary type.
        
        Ex: /edev_12_dstat has an index of 12 so this will return true
        Ex: /edev has no index so this will return false
        """
        return len(self._split) > 1

    def count(self) -> int:
        return len(self._split)

    def join(self, how_many: int) -> str:
        return SEP.join([str(x) for x in self._split[:how_many]])

    def startswith(self, value: str) -> bool:
        return self.href.startswith(value)

    def at(self, index: int) -> Union[str, int, None]:
        try:
            intvalue = int(self._split[index])
            return intvalue
        except ValueError:
            return self._split[index]
        except IndexError:
            return None


class HrefEventParser(HrefParser):

    @property
    def program_index(self) -> int:
        return int(self.at(1))

    @property
    def event_index(self) -> int:
        return int(self._split[-1])

    @property
    def events_href(self) -> str:
        return SEP.join(self._split[:-1])


class EndDeviceHref:

    def __init__(self, index: int = None, edev_href: str = None):
        if index is None and edev_href is None:
            raise ValueError("Must have either index or edev_href specified")

        if index is not None and edev_href is not None:
            raise ValueError("Cannot have both index and edev_href specified")

        self.index = index
        if edev_href is not None:
            self.index = int(edev_href.split(SEP)[1])

        self._root = SEP.join([DEFAULT_EDEV_ROOT, str(self.index)])

    @staticmethod
    def parse(href: str) -> EndDeviceHref:
        index = int(href.split(SEP)[1])
        return EndDeviceHref(index)

    def __str__(self) -> str:
        return SEP.join([DEFAULT_EDEV_ROOT, str(self.index)])

    @property
    def configuration(self) -> str:
        return SEP.join([DEFAULT_EDEV_ROOT, str(self.index), "cfg"])

    @property
    def der_list(self) -> str:
        return SEP.join([DEFAULT_EDEV_ROOT, str(self.index), DER])

    @property
    def device_information(self) -> str:
        return SEP.join([DEFAULT_EDEV_ROOT, str(self.index), END_DEVICE_INFORMATION])

    @property
    def device_status(self) -> str:
        return SEP.join([DEFAULT_EDEV_ROOT, str(self.index), END_DEVICE_STATUS])

    @property
    def power_status(self) -> str:
        return SEP.join([DEFAULT_EDEV_ROOT, str(self.index), END_DEVICE_POWER_STATUS])

    @property
    def registration(self) -> str:
        return SEP.join([DEFAULT_EDEV_ROOT, str(self.index), END_DEVICE_REGISTRATION])

    @property
    def function_set_assignments(self) -> str:
        return SEP.join([DEFAULT_EDEV_ROOT, str(self.index), END_DEVICE_FSA])

    @property
    def log_event_list(self) -> str:
        return SEP.join([DEFAULT_EDEV_ROOT, str(self.index), END_DEVICE_LOG_EVENT_LIST])

    def fill_hrefs(self, enddevice: m.EndDevice) -> m.EndDevice:
        enddevice.href = self._root
        enddevice.ConfigurationLink = m.ConfigurationLink(self.configuration)
        enddevice.DeviceInformationLink = m.DeviceInformationLink(self.device_information)
        enddevice.DeviceStatusLink = m.DeviceStatusLink(self.device_status)
        enddevice.PowerStatusLink = m.PowerStatusLink(self.power_status)
        enddevice.RegistrationLink = m.RegistrationLink(self.registration)
        enddevice.FunctionSetAssignmentsListLink = m.FunctionSetAssignmentsListLink(
            self.function_set_assignments, all=0)
        enddevice.LogEventListLink = m.LogEventListLink(self.log_event_list, all=0)
        enddevice.DERListLink = m.DERListLink(self.der_list, all=0)
        return enddevice


# class DeviceCapabilityHref:

#     def __init__(self, index: int = None, enddevice: EndDeviceHref = None):
#         if index is None and enddevice is None:
#             raise ValueError(f"index and enddevice cannot both be None")
#         elif index is not None and enddevice is not None:
#             if enddevice.index != index:
#                 raise ValueError(f"index and enddevice.index must match if both are specified")

#         if enddevice is not None:
#             index = enddevice.index

#         self.index = index

#     def __str__(self) -> str:
#         return SEP.join([DEFAULT_DCAP_ROOT, str(self.index)])


class DERSubType(Enum):
    Capability = DER_CAPABILITY
    Settings = DER_SETTINGS
    Status = DER_STATUS
    Availability = DER_AVAILABILITY
    CurrentProgram = DER_PROGRAM
    None_Available = NO_INDEX


class FSASubType(Enum):
    DERProgram = "derp"


class DERProgramSubType(Enum):
    NoLink = 0
    ActiveDERControlListLink = 1
    DefaultDERControlLink = 2
    DERControlListLink = 3
    DERCurveListLink = 4
    DERControlReplyTo = 5
    DERControl = 6


class DERHref:

    def __init__(self, root: str) -> None:
        """Constructs a DERHref.
        
        The root should be a single instance not a list.  The properties
        on this object will be the href of the link to the resourse.
        """
        self.root = root

    @property
    def der_availability(self) -> str:
        return SEP.join([self.root, DER_AVAILABILITY])

    @property
    def der_status(self) -> str:
        return SEP.join([self.root, DER_STATUS])

    @property
    def der_capability(self) -> str:
        return SEP.join([self.root, DER_CAPABILITY])

    @property
    def der_settings(self) -> str:
        return SEP.join([self.root, DER_SETTINGS])

    @property
    def der_current_program(self) -> str:
        return SEP.join([self.root, DER_PROGRAM])

    def fill_hrefs(self, der: m.DER) -> m.DER:
        der.href = self.root
        der.DERAvailabilityLink = m.DERAvailabilityLink(self.der_availability)
        der.DERStatusLink = m.DERStatusLink(self.der_status)
        der.DERCapabilityLink = m.DERCapabilityLink(self.der_capability)
        der.DERSettingsLink = m.DERSettingsLink(self.der_settings)
        der.DERProgramLink = m.DERProgramLink(self.der_current_program)
        return der


class DeviceCapabilityHref:

    def __init__(self, end_device_index: str) -> None:
        self._end_device_index = end_device_index
        self.root = DEFAULT_DCAP_ROOT
        #SEP.join([DEFAULT_EDEV_ROOT, self._end_device_index, DER, DER_CAPABILITY])
        #m.DeviceCapability

    @property
    def enddevice_href(self) -> str:
        return DEFAULT_EDEV_ROOT

    @property
    def mirror_usage_point_href(self) -> str:
        return DEFAULT_MUP_ROOT

    @property
    def self_device_href(self) -> str:
        return DEFAULT_SELF_ROOT

    @property
    def time_href(self) -> str:
        return DEFAULT_TIME_ROOT

    @property
    def usage_point_href(self) -> str:
        return DEFAULT_UPT_ROOT

    def fill_hrefs(self, dcap: m.DeviceCapability):
        dcap.href = self.root
        dcap.EndDeviceListLink = m.EndDeviceListLink(self.enddevice_href, all=1)
        dcap.MirrorUsagePointLink = m.MirrorUsagePointListLink(self.mirror_usage_point_href, all=0)
        dcap.SelfDeviceLink = m.SelfDeviceLink(self.self_device_href)
        dcap.TimeLink = m.TimeLink(self.time_href)
        dcap.UsagePointListLink = m.UsagePointListLink(self.usage_point_href, all=0)
        dcap.DERProgramListLink = m.DERProgramListLink(href=DEFAULT_DERP_ROOT, all=0)
        return dcap


class DERProgramHref:

    def __init__(self, program_index: int) -> None:
        self._root = SEP.join([DEFAULT_DERP_ROOT, str(program_index)])

    @property
    def active_control_href(self) -> str:
        return SEP.join([self._root, DER_CONTROL_ACTIVE])

    @property
    def default_control_href(self) -> str:
        return SEP.join([self._root, DDERC])

    @property
    def der_control_list_href(self) -> str:
        return SEP.join([self._root, DERC])

    @property
    def der_curve_list_href(self) -> str:
        return SEP.join([self._root, CURVE])

    def fill_hrefs(self, program: m.DERProgram) -> m.DERProgram:
        program.href = self._root
        program.ActiveDERControlListLink = m.ActiveDERControlListLink(self.active_control_href,
                                                                      all=0)
        program.DefaultDERControlLink = m.DefaultDERControlLink(self.default_control_href)
        program.DERControlListLink = m.DERControlListLink(href=self.der_control_list_href, all=0)
        program.DERCurveListLink = m.DERCurveListLink(href=self.der_curve_list_href, all=0)
        return program


class DERProgramHrefOld(NamedTuple):
    root: str
    index: int
    derp_subtype: DERProgramSubType = DERProgramSubType.NoLink
    derp_subtype_index: int = NO_INDEX

    @staticmethod
    def parse(href: str) -> DERProgramHrefOld:
        parsed = href.split(SEP)
        if len(parsed) == 1:
            return DERProgramHrefOld(parsed[0], NO_INDEX)
        elif len(parsed) == 2:
            return DERProgramHrefOld(parsed[0], int(parsed[1]))
        else:
            mapped = dict(
                derc=DERProgramSubType.DERControlListLink,
                derca=DERProgramSubType.ActiveDERControlListLink,
                dderc=DERProgramSubType.DefaultDERControlLink,
            )
            if len(parsed) == 4:
                return DERProgramHrefOld(parsed[0], int(parsed[1]), mapped[parsed[2]],
                                         int(parsed[3]))
            return DERProgramHrefOld(parsed[0], int(parsed[1]), mapped[parsed[2]])


def der_program_parse(href: str) -> DERProgramHrefOld:
    return DERProgramHrefOld.parse(href)


def der_program_href(index: int = NO_INDEX,
                     sub: DERProgramSubType = DERProgramSubType.NoLink,
                     subindex: int = NO_INDEX) -> str:
    if index == NO_INDEX:
        return DEFAULT_DERP_ROOT

    if sub == DERProgramSubType.NoLink:
        return SEP.join([DEFAULT_DERP_ROOT, str(index)])

    if sub == DERProgramSubType.ActiveDERControlListLink:
        if subindex == NO_INDEX:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DER_CONTROL_ACTIVE])
        else:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DER_CONTROL_ACTIVE, str(subindex)])

    if sub == DERProgramSubType.DefaultDERControlLink:
        if subindex == NO_INDEX:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DDERC])
        else:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DDERC, str(subindex)])

    if sub == DERProgramSubType.DERCurveListLink:
        if subindex == NO_INDEX:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), CURVE])
        else:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), CURVE, str(subindex)])

    if sub == DERProgramSubType.DERControlListLink:
        if subindex == NO_INDEX:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DERC])
        else:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DERC, str(subindex)])

    if sub == DERProgramSubType.DERControlReplyTo:
        return DEFAULT_RSPS_ROOT


@lru_cache()
def get_server_config_href() -> str:
    return "/server/cfg"


@lru_cache()
def get_enddevice_list_href() -> str:
    return DEFAULT_EDEV_ROOT


@lru_cache()
def curve_href(index: int = NO_INDEX) -> str:
    if index == NO_INDEX:
        return DEFAULT_CURVE_ROOT

    return SEP.join([DEFAULT_CURVE_ROOT, str(index)])


@lru_cache()
def fsa_href(index: int = NO_INDEX, edev_index: int = NO_INDEX):
    if index == NO_INDEX and edev_index == NO_INDEX:
        return DEFAULT_FSA_ROOT
    elif index != NO_INDEX and edev_index == NO_INDEX:
        return SEP.join([DEFAULT_FSA_ROOT, str(index)])
    elif index == NO_INDEX and edev_index != NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), FSA])
    else:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), FSA, str(index)])


def derp_href(edev_index: int, fsa_index: int) -> str:
    return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), FSA, str(fsa_index), DER_PROGRAM])


def der_href(index: int = NO_INDEX, fsa_index: int = NO_INDEX, edev_index: int = NO_INDEX):
    if index == NO_INDEX and fsa_index == NO_INDEX and edev_index == NO_INDEX:
        return DEFAULT_DER_ROOT
    elif index != NO_INDEX and fsa_index == NO_INDEX and edev_index == NO_INDEX:
        return SEP.join([DEFAULT_DER_ROOT, str(index)])
    elif index == NO_INDEX and fsa_index != NO_INDEX and edev_index == NO_INDEX:
        return SEP.join([DEFAULT_FSA_ROOT, str(fsa_index), DER_PROGRAM])
    elif edev_index != NO_INDEX and fsa_index == NO_INDEX and index == NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, int(edev_index), FSA])
    elif edev_index != NO_INDEX and fsa_index != NO_INDEX and index == NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, int(edev_index), FSA, int(fsa_index)])
    else:
        raise ValueError(f"index={index}, fsa_index={fsa_index}, edev_index={edev_index}")


def edev_der_href(edev_index: int, der_index: int = NO_INDEX) -> str:
    if der_index == NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER])
    return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER, str(der_index)])


class EDevSubType(Enum):
    None_Available = NO_INDEX
    Registration = END_DEVICE_REGISTRATION
    DeviceStatus = END_DEVICE_STATUS
    PowerStatus = END_DEVICE_POWER_STATUS
    FunctionSetAssignments = END_DEVICE_FSA
    LogEventList = END_DEVICE_LOG_EVENT_LIST
    DeviceInformation = END_DEVICE_INFORMATION
    DER = DER


@dataclass
class EdevHref:
    edev_index: int
    edev_subtype: EDevSubType = EDevSubType.None_Available
    edev_subtype_index: int = NO_INDEX
    edev_der_subtype: DERSubType = DERSubType.None_Available

    def __str__(self) -> str:
        value = "/edev"
        if self.edev_index != NO_INDEX:
            value = f"{value}{SEP}{self.edev_index}"

        if self.edev_subtype != EDevSubType.None_Available:
            value = f"{value}{SEP}{self.edev_subtype.value}"

        if self.edev_subtype_index != NO_INDEX:
            value = f"{value}{SEP}{self.edev_subtype_index}"

        if self.edev_der_subtype != DERSubType.None_Available:
            value = f"{value}{SEP}{self.edev_der_subtype.value}"

        return value

    def parse(path: str) -> EdevHref:
        split_pth = path.split(SEP)

        if split_pth[0] != EDEV and split_pth[0][1:] != EDEV:
            raise ValueError(f"Must start with {EDEV}")

        if len(split_pth) == 1:
            return EdevHref(NO_INDEX)
        elif len(split_pth) == 2:
            return EdevHref(int(split_pth[1]))
        elif len(split_pth) == 3:
            return EdevHref(int(split_pth[1]), edev_subtype=EDevSubType(split_pth[2]))
        elif len(split_pth) == 4:
            return EdevHref(int(split_pth[1]),
                            edev_subtype=EDevSubType(split_pth[2]),
                            edev_subtype_index=int(split_pth[3]))
        elif len(split_pth) == 5:
            return EdevHref(int(split_pth[1]),
                            edev_subtype=EDevSubType(split_pth[2]),
                            edev_subtype_index=int(split_pth[3]),
                            edev_der_subtype=DERSubType(split_pth[4]))
        else:
            raise ValueError("Out of bounds parsing.")

    def __eq__(self, other: object) -> bool:
        return other.edev_index == self.edev_index and other.edev_subtype == self.edev_subtype, \
            other.edev_subtype_index == self.edev_subtype_index and other.edev_der_subtype == self.edev_der_subtype


class FSAHref(NamedTuple):
    fsa_index: NO_INDEX
    fsa_sub: FSASubType = None


def fsa_parse(path: str) -> FSAHref:
    split_pth = path.split(SEP)

    if len(split_pth) == 1:
        return FSAHref(NO_INDEX)
    elif len(split_pth) == 2:
        return FSAHref(int(split_pth[1]))
    elif len(split_pth) == 3:
        return FSAHref(int(split_pth[1]), fsa_sub=split_pth[2])

    raise ValueError("Invalid parsing path.")


def der_sub_href(edev_index: int, index: int = NO_INDEX, subtype: DERSubType = None):
    if subtype is None and index == NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER])
    elif subtype is None:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER, str(index)])
    else:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER, str(index), subtype.value])


@lru_cache()
def mirror_usage_point_href(mirror_usage_point_index: int = NO_INDEX):
    """Mirror Usage Point hrefs
    
       /mup
       /mup/{mirror_usage_point_index}
       
    
    """
    if mirror_usage_point_index == NO_INDEX:
        ret = DEFAULT_MUP_ROOT
    else:
        ret = SEP.join([DEFAULT_MUP_ROOT, str(mirror_usage_point_index)])

    return ret


class ParsedUsagePointHref:

    def __init__(self, href: str):
        self._href = href
        self._split = href.split(SEP)

    def last_list(self) -> str:
        """Assuming the parsed href has a reference to an item, return the container href.
        """
        if self._split[-1].isnumeric():
            return SEP.join(self._split[:-1])
        return self._href

    def has_usage_point_index(self) -> bool:
        return self.usage_point_index is not None

    def has_extra(self) -> bool:
        return self.has_meter_reading_list() or \
                self.has_reading_list() or \
                self.has_reading_set_list() or \
                self.has_reading_set_reading_list()

    def has_meter_reading_list(self) -> bool:
        try:
            if retval := self._split[2] == "mr":
                ...
            return retval
        except IndexError:
            return False

    def has_reading_type(self) -> bool:
        try:
            if retval := self._split[4] == "rt":
                ...
            return retval
        except IndexError:
            return False

    def has_reading_set_list(self) -> bool:
        try:
            if retval := self._split[4] == "rs":
                ...
            return retval
        except IndexError:
            return False

    def has_reading_set_reading_list(self) -> bool:
        try:
            if retval := self._split[6] == "r":
                ...
            return retval
        except IndexError:
            return False

    def has_reading_list(self) -> bool:
        try:
            if retval := self._split[4] == "r" or self._split[6] == "r":
                ...
            return retval
        except IndexError:
            return False

    @property
    def usage_point_index(self) -> Optional[int]:
        try:
            return int(self._split[1])
        except IndexError:
            pass
        return None

    @property
    def meter_reading_index(self) -> Optional[int]:
        try:
            return int(self._split[3])
        except IndexError:
            pass
        return None

    @property
    def reading_set_index(self) -> Optional[int]:
        try:
            if self._split[4] == "rs":
                return int(self._split[5])
        except IndexError:
            pass
        return None

    @property
    def reading_set_reading_index(self) -> Optional[int]:
        try:
            if self._split[6] == "r":
                return int(self._split[7])
        except IndexError:
            pass
        return None

    @property
    def reading_index(self) -> Optional[int]:
        try:
            if self._split[4] == "r":
                return int(self._split[5])
            elif self._split[6] == "r":
                return int(self._split[7])
        except IndexError:
            pass

        return None


class UsagePointHref:

    def __init__(self, href: str = None, root: str = '/upt'):
        self._href = href
        self._root = root

    def is_root(self) -> bool:
        return self._href == self._root

    def value(self) -> str:
        return self._root

    def usage_point(self, usage_point_index: int):
        return SEP.join([self._root, str(usage_point_index)])

    def meterreading_list(self, usage_point_index: int) -> str:
        return SEP.join([self._root, str(usage_point_index), "mr"])

    def meterreading(self, usage_point_index: int, meter_reading_index: int) -> str:
        return SEP.join([self.meterreading_list(usage_point_index), str(meter_reading_index)])

    def readingset_list(self, usage_point_index: int, meter_reading_index: int) -> str:
        return SEP.join([self.meterreading(usage_point_index, meter_reading_index), "rs"])

    def readingtype(self, usage_point_index: int, meter_reading_index: int) -> str:
        return SEP.join([self.meterreading(usage_point_index, meter_reading_index), "rt"])

    def readingset(self, usage_point_index: int, meter_reading_index: int,
                   reading_set_index: int) -> str:
        return SEP.join(
            [self.readingset_list(usage_point_index, meter_reading_index),
             str(reading_set_index)])

    def readingsetreading_list(self, usage_point_index: int, meter_reading_index: int,
                               reading_set_index: int):
        return SEP.join(
            [self.readingset(usage_point_index, meter_reading_index, reading_set_index), "r"])

    def readingsetreading(self, usage_point_index: int, meter_reading_index: int,
                          reading_set_index: int, reading_index: int):
        return SEP.join([
            self.readingsetreading_list(usage_point_index, meter_reading_index, reading_set_index),
            str(reading_index)
        ])

    def reading_list(self, usage_point_index: int, meter_reading_index: int) -> str:
        return SEP.join([self.meterreading(usage_point_index, meter_reading_index), "r"])

    def reading(self, usage_point_index: int, meter_reading_index: int, reading_index: int) -> str:
        return SEP.join(
            [self.reading_list(usage_point_index, meter_reading_index),
             str(reading_index)])


@dataclass
class MirrorUsagePointHref:
    mirror_usage_point_index: int = NO_INDEX
    meter_reading_list_index: int = NO_INDEX
    meter_reading_index: int = NO_INDEX
    reading_set_index: int = NO_INDEX
    reading_index: int = NO_INDEX

    @staticmethod
    def parse(href: str) -> MirrorUsagePointHref:
        items = href.split(SEP)
        if len(items) == 1:
            return MirrorUsagePointHref()

        if len(items) == 2:
            return MirrorUsagePointHref(items[1])


def usage_point_href(usage_point_index: int | str = NO_INDEX,
                     meter_reading_list: bool = False,
                     meter_reading_list_index: int = NO_INDEX,
                     meter_reading_index: int = NO_INDEX,
                     meter_reading_type: bool = False,
                     reading_set: bool = False,
                     reading_set_index: int = NO_INDEX,
                     reading_index: int = NO_INDEX):
    """Usage point hrefs 

       /upt
       /upt/{usage_point_index}
       /upt/{usage_point_index}/mr
       /upt/{usage_point_index}/mr/{meter_reading_index}
       /upt/{usage_point_index}/mr/{meter_reading_index}/rt
       /upt/{usage_point_index}/mr/{meter_reading_index}/rs
       /upt/{usage_point_index}/mr/{meter_reading_index}/rs/{reading_set_index}
       /upt/{usage_point_index}/mr/{meter_reading_index}/rs/{reading_set_index}/r
       /upt/{usage_point_index}/mr/{meter_reading_index}/rs/{reading_set_index}/r/{reading_index}
       
       

    """
    if isinstance(usage_point_index, str):
        base_upt = usage_point_index
    else:
        base_upt = DEFAULT_UPT_ROOT

    if usage_point_index == NO_INDEX:
        ret = base_upt
    else:
        if isinstance(usage_point_index, str):
            arr = [base_upt]
        else:
            arr = [DEFAULT_UPT_ROOT, str(usage_point_index)]

        if meter_reading_list:
            if meter_reading_list_index == NO_INDEX:
                arr.extend(["mr"])
            else:
                arr.extend(["mr", str(meter_reading_list_index)])

        ret = SEP.join(arr)
    return ret


def get_der_program_list(fsa_href: str) -> str:
    return SEP.join([fsa_href, "der"])


def get_dr_program_list(fsa_href: str) -> str:
    return SEP.join([fsa_href, "dr"])


def get_fsa_list_href(end_device_href: str) -> str:
    return SEP.join([end_device_href, "fsa"])


def get_response_set_href():
    return DEFAULT_RSPS_ROOT


@lru_cache()
def get_der_list_href(index: int) -> str:
    if index == NO_INDEX:
        ret = DEFAULT_DER_ROOT
    else:
        ret = SEP.join([DEFAULT_DER_ROOT, str(index)])
    return ret


@lru_cache()
def get_enddevice_href(edev_indx: int = NO_INDEX, subref: str = None) -> str:
    if edev_indx == NO_INDEX:
        ret = DEFAULT_EDEV_ROOT
    elif subref:
        ret = SEP.join([DEFAULT_EDEV_ROOT, f"{edev_indx}", f"{subref}"])
    else:
        ret = SEP.join([DEFAULT_EDEV_ROOT, f"{edev_indx}"])
    return ret


@lru_cache()
def registration_href(edev_index: int) -> str:
    return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), "rg"])


@lru_cache()
def get_configuration_href(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "cfg")


@lru_cache()
def get_power_status_href(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "ps")


@lru_cache()
def get_device_status(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "ds")


@lru_cache()
def get_device_information(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "di")


@lru_cache()
def get_time_href() -> str:
    # return f"{DEFAULT_DCAP_ROOT}{SEP}tm"
    return f"/tm"


@lru_cache()
def get_log_list_href(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "lel")


@lru_cache()
def get_dcap_href() -> str:
    return f"{DEFAULT_DCAP_ROOT}"


def get_dderc_href() -> str:
    return SEP.join([DEFAULT_DER_ROOT, DDERC])


def get_derc_default_href(derp_index: int) -> str:
    return SEP.join([DEFAULT_DER_ROOT, DDERC, f"{derp_index}"])


def get_derc_href(index: int) -> str:
    """Return the DERControl href to the caller

    if NO_INDEX then don't include the index in the result.
    """
    if index == NO_INDEX:
        return SEP.join([DEFAULT_DER_ROOT, DERC])

    return SEP.join([DEFAULT_DER_ROOT, DERC, f"{index}"])


def get_program_href(index: int, subref: str = None):
    """Return the DERProgram href to the caller

    Args:
        index: if NO_INDEX then don't include the index in the result else use the index
        subref: used to specify a subsection in the program.
    """
    if index == NO_INDEX:
        ref = f"{DEFAULT_DERP_ROOT}"
    else:
        if subref is not None:
            ref = f"{DEFAULT_DERP_ROOT}{SEP}{index}{SEP}{subref}"
        else:
            ref = f"{DEFAULT_DERP_ROOT}{SEP}{index}"
    return ref


sdev: str = DEFAULT_SELF_ROOT

admin: str = "/admin"
uuid_gen: str = "/uuid"


def build_link(base_url: str, *suffix: Optional[str]):
    result = base_url
    if result.endswith("/"):
        result = result[:-1]

    if suffix:
        for p in suffix:
            if p is not None:
                if isinstance(p, str):
                    if p.startswith("/"):
                        result += f"{p}"
                    else:
                        result += f"/{p}"
                else:
                    result += f"/{p}"

    return result


def extend_url(base_url: str, index: Optional[int] = None, suffix: Optional[str] = None):
    result = base_url
    if index is not None:
        result += f"/{index}"
    if suffix:
        result += f"/{suffix}"

    return result



import json
import atexit
import os
import sys
import time
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from pprint import pprint, pformat
from queue import Queue
from threading import Thread

from gridappsd.field_interface import MessageBusDefinition, ContextManager
from gridappsd.field_interface.agents import FeederAgent, SecondaryAreaAgent

from Queries import QueryAllDERGroups, QueryBattery, QuerySolar, QueryInverter
# from ieee_2030_5.models import Resource, PowerStatus, DERCapability, UsagePoint
# from ieee_2030_5.models.end_devices import EndDevices
from ieee_2030_5.models import UsagePoint


class DataPumpFeederAgent(FeederAgent):

    def __init__(self, upstream_message_bus_def: MessageBusDefinition,
                 downstream_message_bus_def: MessageBusDefinition = None,
                 feeder_dict=None, simulation_id=None):
        super().__init__(upstream_message_bus_def, downstream_message_bus_def, feeder_dict, simulation_id)
    #TODO remove first four
    def on_measurement(self, peer, sender, bus, topic, headers, message):
        # with open("feeder.txt", "a") as fp:
        #     fp.write(json.dumps(message))
        print("Feeder!")
        print(message)

class DataPumperAgent(SecondaryAreaAgent):
    def __init__(self, upstream_message_bus_def: MessageBusDefinition, downstream_message_bus_def: MessageBusDefinition,
                 secondary_area_dict=None, simulation_id=None):
        super().__init__(upstream_message_bus_def, downstream_message_bus_def, secondary_area_dict, simulation_id)

    def on_measurement(self, peer, sender, bus, topic, headers, message):
        # with open("secondary.txt", "a") as fp:
        #     if "_4c491539-dfc1-4fda-9841-3bf10348e2fa" in json.dumps(message):
        #         print("Woot found it!")
        #         sys.exit()
        #     fp.write(json.dumps(message))
        print("Secondary Area")
        print(message)

def start_data_pump(msg_bus_def: MessageBusDefinition):

    atexit.register(stop_data_pump)

def stop_data_pump():
    pass


if __name__ == '__main__':
    os.environ["GRIDAPPSD_USER"] = "system"
    os.environ["GRIDAPPSD_PASSWORD"] = "manager"
    os.environ["GRIDAPPSD_ADDRESS"] = "gridappsd"
    os.environ["GRIDAPPSD_PORT"] = "61613"

    def run_simulation(queue: Queue):
        import json
        import os
        from pathlib import Path

        from gridappsd import GridAPPSD
        from gridappsd.simulation import Simulation

        import auth_context

        sim_config = json.load(Path("/repos/gridappsd-2030_5/examples/config_files_simulated/simulation-config.json").open())
        # sim_config = json.load(Path("config_files_simulated/simulation-config.json").open())
        sim_feeder = sim_config['power_system_config']['Line_name']
        print(f"Simulation for feeder: {sim_feeder}")
        gapps = GridAPPSD()
        sim = Simulation(gapps, run_config=sim_config)
        print("Starting Simulation")
        sim.start_simulation()
        assert sim.simulation_id
        print(f"Simulation id is {sim.simulation_id}")
        Path("simulation.feeder.txt").write_text(sim_feeder)
        Path("simulation.id.txt").write_text(sim.simulation_id)
        try:
            queue.put("Done")
            sim.run_loop()
        except KeyboardInterrupt:
            print("Stopping simulation")
            sim.stop()

        gapps.disconnect()

    my_wait_queue = Queue()
    sim_thread = Thread(target=run_simulation, args=(my_wait_queue,))
    sim_thread.daemon = True
    sim_thread.start()

    my_wait_queue.get(block=True)
    config_file = """
connections:
  id: _49AD8E07-3BF9-A4E2-CB8F-C3722F837B62
  is_ot_bus: true
  connection_type: CONNECTION_TYPE_GRIDAPPSD
  connection_args:
    GRIDAPPSD_ADDRESS: tcp://gridappsd:61613
    GRIDAPPSD_USER: system
    GRIDAPPSD_PASSWORD: manager
"""
    Path("../server/tmp.file.yml").write_text(config_file)
    system_bus_def = MessageBusDefinition.load("../server/tmp.file.yml")

    feeder_id = Path("simulation.feeder.txt").read_text().strip()
    simulation_id = Path("simulation.id.txt").read_text().strip()
    # feeder_id = "_49AD8E07-3BF9-A4E2-CB8F-C3722F837B62"
    # imulation_id = "1160276262"

    ieee_resources = []
    inverters = QueryInverter(feeder_id)

    context = ContextManager.get_context_by_feeder(feeder_id)
    pprint(context)
    bus_refs = defaultdict(list)

    for switch_area in context['data']['switch_areas']:
        for secondary_area in switch_area['secondary_areas']:
            for binding in inverters['data']['results']['bindings']:

                # TODO: Change to use id later...Topo processor is using pecid for now.
                pprint(binding['pecid']['value'])
                pecid = binding['pecid']['value']
                mrid = binding['id']['value']
                # ieee_resources.append(
                #     UsagePoint(mRID=mrid)
                # )
                # pecid is being found, but id is not
                #ieee_resources.append(UsagePoint(mRID=other_mrid))
                ieee_resources.append(UsagePoint(mRID=pecid))
                if pecid in secondary_area['addressable_equipment']:
                    new_area = deepcopy(secondary_area)
                    new_area['addressable_equipment'] = [pecid, mrid]
                    new_area['unaddressable_equipment'] = []

                    bus_refs[secondary_area['message_bus_id']].append(new_area)

    pprint(bus_refs)
    feeder = context['data']
    Path("../server/data.dump.json").write_text(pformat(context['data'], indent=2))
    data = Path("../server/data.dump.json").read_text()
    for p in ieee_resources:
        if str(p.mRID) in data:
            print(f"Found: {p.mRID}")
        else:
            print(f"Not found: {p.mRID}")
    # feeder_agent = DataPumpFeederAgent(upstream_message_bus_def=system_bus_def,
    #                                    feeder_dict=feeder,
    #                                    simulation_id=simulation_id)
    # feeder_agent.connect()
    for bus_id, secondary_area in bus_refs.items():
        bus_def = deepcopy(system_bus_def)
        bus_def.id = bus_id
        for area in secondary_area:
            dpa = DataPumperAgent(system_bus_def, downstream_message_bus_def=bus_def, secondary_area_dict=area,
                                  simulation_id=simulation_id)
            dpa.connect()

    while True:
        try:
            time.sleep(0.1)
        except KeyboardInterrupt:
            print("Exiting sample")
            break

    sim_thread.join(timeout=5)
    #         break
    # coordinating_agent.spawn_distributed_agent(feeder_agent)
    # addressable = context["addressable_equipment"]
    # print(addressable)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import List, Optional

__NAMESPACE__ = "http://pypi.org/project/xsdata"


@dataclass
class TypeName:

    class Meta:
        name = "ClassName"
        namespace = "http://pypi.org/project/xsdata"

    case: Optional[str] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    safePrefix: Optional[str] = field(default=None,
                                      metadata={
                                          "type": "Attribute",
                                          "required": True,
                                      })


@dataclass
class CompoundFields:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    defaultName: Optional[str] = field(default=None,
                                       metadata={
                                           "type": "Attribute",
                                           "required": True,
                                       })
    forceDefaultName: Optional[bool] = field(default=None,
                                             metadata={
                                                 "type": "Attribute",
                                                 "required": True,
                                             })
    value: Optional[bool] = field(default=None, metadata={
        "required": True,
    })


@dataclass
class ConstantName:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    case: Optional[str] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    safePrefix: Optional[str] = field(default=None,
                                      metadata={
                                          "type": "Attribute",
                                          "required": True,
                                      })


@dataclass
class FieldName:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    case: Optional[str] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    safePrefix: Optional[str] = field(default=None,
                                      metadata={
                                          "type": "Attribute",
                                          "required": True,
                                      })


@dataclass
class Format:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    repr: Optional[bool] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    eq: Optional[bool] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    order: Optional[bool] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    unsafeHash: Optional[bool] = field(default=None,
                                       metadata={
                                           "type": "Attribute",
                                           "required": True,
                                       })
    frozen: Optional[bool] = field(default=None,
                                   metadata={
                                       "type": "Attribute",
                                       "required": True,
                                   })
    slots: Optional[bool] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    kwOnly: Optional[bool] = field(default=None,
                                   metadata={
                                       "type": "Attribute",
                                       "required": True,
                                   })
    value: str = field(default="", metadata={
        "required": True,
    })


@dataclass
class ModuleName:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    case: Optional[str] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    safePrefix: Optional[str] = field(default=None,
                                      metadata={
                                          "type": "Attribute",
                                          "required": True,
                                      })


@dataclass
class PackageName:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    case: Optional[str] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    safePrefix: Optional[str] = field(default=None,
                                      metadata={
                                          "type": "Attribute",
                                          "required": True,
                                      })


@dataclass
class Substitution:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    type_value: Optional[str] = field(default=None,
                                      metadata={
                                          "name": "type",
                                          "type": "Attribute",
                                          "required": True,
                                      })
    search: Optional[str] = field(default=None, metadata={
        "type": "Attribute",
        "required": True,
    })
    replace: Optional[str] = field(default=None,
                                   metadata={
                                       "type": "Attribute",
                                       "required": True,
                                   })


@dataclass
class Conventions:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    ClassName: Optional[TypeName] = field(default=None,
                                          metadata={
                                              "type": "Element",
                                              "required": True,
                                          })
    FieldName: Optional[FieldName] = field(default=None,
                                           metadata={
                                               "type": "Element",
                                               "required": True,
                                           })
    ConstantName: Optional[ConstantName] = field(default=None,
                                                 metadata={
                                                     "type": "Element",
                                                     "required": True,
                                                 })
    ModuleName: Optional[ModuleName] = field(default=None,
                                             metadata={
                                                 "type": "Element",
                                                 "required": True,
                                             })
    PackageName: Optional[PackageName] = field(default=None,
                                               metadata={
                                                   "type": "Element",
                                                   "required": True,
                                               })


@dataclass
class Output:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    maxLineLength: Optional[int] = field(default=None,
                                         metadata={
                                             "type": "Attribute",
                                             "required": True,
                                         })
    Package: Optional[str] = field(default=None, metadata={
        "type": "Element",
        "required": True,
    })
    Format: Optional[Format] = field(default=None,
                                     metadata={
                                         "type": "Element",
                                         "required": True,
                                     })
    Structure: Optional[str] = field(default=None,
                                     metadata={
                                         "type": "Element",
                                         "required": True,
                                     })
    DocstringStyle: Optional[str] = field(default=None,
                                          metadata={
                                              "type": "Element",
                                              "required": True,
                                          })
    FilterStrategy: Optional[str] = field(default=None,
                                          metadata={
                                              "type": "Element",
                                              "required": True,
                                          })
    RelativeImports: Optional[bool] = field(default=None,
                                            metadata={
                                                "type": "Element",
                                                "required": True,
                                            })
    CompoundFields: Optional[CompoundFields] = field(default=None,
                                                     metadata={
                                                         "type": "Element",
                                                         "required": True,
                                                     })
    PostponedAnnotations: Optional[bool] = field(default=None,
                                                 metadata={
                                                     "type": "Element",
                                                     "required": True,
                                                 })
    UnnestClasses: Optional[bool] = field(default=None,
                                          metadata={
                                              "type": "Element",
                                              "required": True,
                                          })
    IgnorePatterns: Optional[bool] = field(default=None,
                                           metadata={
                                               "type": "Element",
                                               "required": True,
                                           })


@dataclass
class Substitutions:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    Substitution: List[Substitution] = field(default_factory=list,
                                             metadata={
                                                 "type": "Element",
                                                 "min_occurs": 1,
                                             })


@dataclass
class Config:

    class Meta:
        namespace = "http://pypi.org/project/xsdata"

    version: Optional[float] = field(default=None,
                                     metadata={
                                         "type": "Attribute",
                                         "required": True,
                                     })
    Output: Optional[Output] = field(default=None,
                                     metadata={
                                         "type": "Element",
                                         "required": True,
                                     })
    Conventions: Optional[Conventions] = field(default=None,
                                               metadata={
                                                   "type": "Element",
                                                   "required": True,
                                               })
    Substitutions: Optional[Substitutions] = field(default=None,
                                                   metadata={
                                                       "type": "Element",
                                                       "required": True,
                                                   })



from __future__ import annotations

import pickle
from copy import deepcopy
from dataclasses import dataclass, field
import logging

from datetime import datetime
from email.utils import format_datetime
from typing import Dict, Optional, List
from ieee_2030_5.models.sep import Link
from ieee_2030_5.persistance.points import set_point, get_point

__all__: List[str] = ["get_href", "add_href", "get_href_all_names", "get_href_filtered"]

_log = logging.getLogger(__name__)


@dataclass
class Index:
    href: str
    item: object
    added: str    # Optional[Union[datetime | str]]
    last_written: str    # Optional[Union[datetime | str]]
    last_hash: Optional[int]


@dataclass
class Indexer:
    __items__: Dict = field(default=None)

    def init(self):
        if self.__items__ is None:
            self.__items__ = {}

    @property
    def length(self) -> int:
        self.init()
        return len(self.__items__)

    def add(self, href: str, item: dataclass):
        self.init()

        # TODO: Verify that this method actually works with a new object.
        # If using a link, we need the true href to cache the object.
        if isinstance(href, Link):
            href = href.href
        # cached = self.__items__.get(href)
        # if cached and cached.item == item:
        #     _log.debug(f"Item already cached {href}")
        # else:
        added = format_datetime(datetime.utcnow())
        serialized_item = pickle.dumps(
            item)    # serialize_dataclass(item, serialization_type=SerializeType.JSON)
        obj = Index(href, item, added=added, last_written=added, last_hash=hash(serialized_item))
        # serialized_obj = serialize_dataclass(obj, serialization_type=SerializeType.JSON)

        # note storing Index object.
        set_point(href, pickle.dumps(
            obj))    # serialize_dataclass(obj, serialization_type=SerializeType.JSON))
        self.__items__[href] = obj

    def get(self, href) -> dataclass:
        self.init()
        # If using a link, we need the true href to cache the object.
        if isinstance(href, Link):
            href = href.href
        if href in self.__items__:
            index = pickle.loads(get_point(href))    # pickle.loads(get_point(href))
            # index = pickle.loads(self.__items__.get(href))
            # index = deserialize_dataclass(data, SerializeType.JSON)
            data = index.item
        else:
            data = None

        return data

    def get_all(self) -> List:
        return deepcopy([x.item for x in self.__items__.values()])


__indexer__ = Indexer()


def add_href(href: str, item: dataclass):
    __indexer__.add(href, item)


def get_href(href: str) -> dataclass:
    return __indexer__.get(href)


def get_href_filtered(href_prefix: str) -> List[dataclass] | []:
    if __indexer__.__items__ is None:
        return []

    return [
        v.item for k, v in __indexer__.__items__.items()
        if k.startswith(href_prefix) and v.item is not None
    ]


def get_href_all_names():
    return [x for x in __indexer__.__items__.keys() if __indexer__.__items__[x] is not None]



from flask import Response

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
from ieee_2030_5.server.base_request import RequestOp


class Dcap(RequestOp):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        # TODO: Test for allowed dcap here.
        # if not self._end_devices.allowed_to_connect(self.lfdi):
        #     raise werkzeug.exceptions.Unauthorized()
        device = adpt.EndDeviceAdapter.fetch_by_property("lFDI", self.lfdi)
        device_index = adpt.EndDeviceAdapter.fetch_index(device)
        cap = adpt.DeviceCapabilityAdapter.fetch(device_index)

        return self.build_response_from_dataclass(cap)


import hashlib
import itertools
import json
import logging
import os
import ssl
import threading
import time
from dataclasses import fields
from functools import lru_cache
from pathlib import Path
from queue import Queue

import OpenSSL
import werkzeug.exceptions
from flask import (Flask, Response, g, redirect, render_template, request, url_for)
from flask_session import Session
# from flask_socketio import SocketIO, send
from werkzeug.serving import BaseWSGIServer, make_server

from ieee_2030_5.utils import dataclass_to_xml

__all__ = ["build_server"]

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.certs import (TLSRepository, lfdi_from_fingerprint, sfdi_from_lfdi)
# templates = Jinja2Templates(directory="templates")
from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.data.indexer import get_href, get_href_all_names
from ieee_2030_5.models import DeviceCategoryType
from ieee_2030_5.server.admin_endpoints import AdminEndpoints
#from ieee_2030_5.server.server_constructs import EndDevices, get_groups
from ieee_2030_5.server.server_endpoints import ServerEndpoints

_log = logging.getLogger(__file__)

server_config: ServerConfiguration = None
tls_repository: TLSRepository = None


class LogProtocolHandler(logging.Handler):

    def __init__(self, level=0) -> None:
        super().__init__(level)
        super().setFormatter('%(message)s')

    def emit(self, record: logging.LogRecord) -> None:
        pass    # super().handle(record)


_log_protocol = logging.getLogger("protocol")
_log_protocol.addHandler(LogProtocolHandler(logging.DEBUG))


class PeerCertWSGIRequestHandler(werkzeug.serving.WSGIRequestHandler):
    """
    We subclass this class so that we can gain access to the connection
    property. self.connection is the underlying client socket. When a TLS
    connection is established, the underlying socket is an instance of
    SSLSocket, which in turn exposes the getpeercert() method.

    The output from that method is what we want to make available elsewhere
    in the application.
    """
    config: ServerConfiguration
    tlsrepo: TLSRepository
    reqresponse: Queue()

    @staticmethod
    @lru_cache
    def is_admin(path_info) -> bool:
        start_paths = ['/admin', '/socket-io']
        a_filter = itertools.accumulate(start_paths,
                                        lambda x: 1 if path_info.startswith(x) else 0,
                                        initial=0)
        return next(a_filter) > 0

    def make_environ(self):
        """
        The superclass method develops the environ hash that eventually
        forms part of the Flask request object.

        We allow the superclass method to run first, then we insert the
        peer certificate into the hash. That exposes it to us later in
        the request variable that Flask provides
        """
        _log.debug("Making environment")
        environ = super(PeerCertWSGIRequestHandler, self).make_environ()

        # Assume browser is being hit with things that start with /admin allow
        # a pass through from web (should be protected via auth but not right now)
        if PeerCertWSGIRequestHandler.is_admin(
                environ['PATH_INFO']) and not self.config.generate_admin_cert:
            raise werkzeug.exceptions.Forbidden()

        try:
            # Short circuit the connection from the client and utilize http rather
            # than x509 certificates.
            if self.config.lfdi_client:
                environ['ieee_2030_5_lfdi'] = self.config.lfdi_client
                environ['ieee_2030_5_sfdi'] = sfdi_from_lfdi(self.config.lfdi_client)
                return environ

            # For admin use the admin peer even though it's not what is sent in to the client.
            # This allows admin to login from any api, though not necessarily secure this
            # allows a way to have the admin be boxed off.
            if PeerCertWSGIRequestHandler.is_admin(environ['PATH_INFO']):
                cert, key = self.tlsrepo.get_file_pair("admin")
                x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, cert)
            else:
                x509_binary = self.connection.getpeercert(True)
                x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_ASN1, x509_binary)
            environ['ieee_2030_5_peercert'] = x509
            environ['ieee_2030_5_serial_number'] = x509.get_serial_number()
            if PeerCertWSGIRequestHandler.config.lfdi_mode == "lfdi_mode_from_file":
                _log.debug("Using hash from combined file.")
                pth = PeerCertWSGIRequestHandler.tlsrepo.__get_combined_file__(
                    x509.get_subject().CN)
                sha256hash = hashlib.sha256(pth.read_text().encode('utf-8')).hexdigest()
                environ['ieee_2030_5_lfdi'] = lfdi_from_fingerprint(sha256hash)
            else:
                environ['ieee_2030_5_lfdi'] = lfdi_from_fingerprint(
                    x509.digest("sha256").decode('ascii'))
            environ['ieee_2030_5_sfdi'] = sfdi_from_lfdi(environ['ieee_2030_5_lfdi'])

            _log.debug(
                f"Environment lfdi: {environ['ieee_2030_5_lfdi']} sfdi: {environ['ieee_2030_5_sfdi']}"
            )
            if not PeerCertWSGIRequestHandler.is_admin(environ['PATH_INFO']):
                # TODO Currently if we are in full file mode there isn't a way to verify that the
                # device is known.
                if PeerCertWSGIRequestHandler.config.lfdi_mode == "lfdi_mode_from_cert_fingerprint":
                    found_device_id = self.tlsrepo.find_device_id_from_sfdi(
                        environ['ieee_2030_5_sfdi'])
                    assert found_device_id, "Unknown device found."

        except OpenSSL.crypto.Error:
            # Only if we have a debug_device do we want to expose this device through the admin page.
            # if self.debug_device:
            #     cert_file, key_file = self.tlsrepo.get_file_pair(self.debug_device)
            #     x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, Path(cert_file).read_bytes())
            #     environ['ieee_2030_5_peercert'] = x509
            #     environ['ieee_2030_5_subject'] = x509.get_subject().CN

            # else:
            environ['peercert'] = None

        return environ


# based on
# https://stackoverflow.com/questions/19459236/how-to-handle-413-request-entity-too-large-in-python-flask-server#:~:text=server%20MAY%20close%20the%20connection,client%20from%20continuing%20the%20request.&text=time%20the%20client%20MAY%20try,you%20the%20Broken%20pipe%20error.&text=Great%20than%20the%20application%20is%20acting%20correct.
def handle_chunking():
    """
    Sets the "wsgi.input_terminated" environment flag, thus enabling
    Werkzeug to pass chunked requests as streams.  The gunicorn server
    should set this, but it's not yet been implemented.
    """

    transfer_encoding = request.headers.get("Transfer-Encoding", None)
    if transfer_encoding == u"chunked":
        request.environ["wsgi.input_terminated"] = True


def before_request():

    g.SERVER_CONFIG = server_config
    g.TLS_REPOSITORY = tls_repository
    # if request.scheme == 'http':
    #     url = request.url.replace('http://', 'https://', 1)
    #     url = url.replace('8080', '7443')
    #     code = 301
    #     return redirect(url, code=code)
    #session['SERVER_CONFIG'] = server_config
    #PeerCertWSGIRequestHandler.reqresponse.put(request)
    # _log.debug(f"HEADERS: {request.headers}")
    # _log.debug(f"REQ_path: {request.path}")
    # _log.debug(f"ARGS: {request.args}")
    # _log.debug(f"DATA: {request.get_data()}")
    # _log.debug(f"FORM: {request.form}")


def after_request(response: Response) -> Response:
    _log_protocol.debug(f"\nREQ: {request.path}")
    _log_protocol.debug(f"\nRESP HEADER: {str(response.headers).strip()}")
    first_line = response.get_data().decode('utf-8').split('\n')[0]
    _log_protocol.debug(f"\nRESP: {first_line}")

    # _log.debug(f"RESP HEADERS:\n{response.headers}")
    # _log.debug(f"RESP:\n{response.get_data().decode('utf-8')}")
    return response


def __build_ssl_context__(tlsrepo: TLSRepository) -> ssl.SSLContext:
    # to establish an SSL socket we need the private key and certificate that
    # we want to serve to users.
    server_key_file = str(tlsrepo.server_key_file)
    server_cert_file = str(tlsrepo.server_cert_file)

    # in order to verify client certificates we need the certificate of the
    # CA that issued the client's certificate. In this example I have a
    # single certificate, but this could also be a bundle file.
    ca_cert = str(tlsrepo.ca_cert_file)

    # create_default_context establishes a new SSLContext object that
    # aligns with the purpose we provide as an argument. Here we provide
    # Purpose.CLIENT_AUTH, so the SSLContext is set up to handle validation
    # of client certificates.
    ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH, cafile=str(ca_cert))

    # load in the certificate and private key for our server to provide to clients.
    # force the client to provide a certificate.
    ssl_context.load_cert_chain(
        certfile=server_cert_file,
        keyfile=server_key_file,
    # password=app_key_password
    )
    # change this to ssl.CERT_REQUIRED during deployment.
    # TODO if required we have to have one all the time on the server.
    ssl_context.verify_mode = ssl.CERT_OPTIONAL    # ssl.CERT_REQUIRED
    return ssl_context


def __build_http_app__(config: ServerConfiguration) -> Flask:
    app = Flask(__name__, template_folder=str(Path(".").resolve().joinpath('templates')))
    # Debug headers path and request arguments
    app.before_request(before_request)
    # Allows for larger data to be sent through because of chunking types.
    app.before_request(handle_chunking)
    app.after_request(after_request)

    @app.route("/dcap", methods=['GET'])
    def http_root() -> Response:
        return dataclass_to_xml(m.DeviceCapability(href=f"https://localhost:7443/dcap"))
        #return adpt.DeviceCapabilityAdapter()


def __build_app__(config: ServerConfiguration, tlsrepo: TLSRepository) -> Flask:
    app = Flask(__name__, template_folder=str(Path(".").resolve().joinpath('templates')))

    # Debug headers path and request arguments
    app.before_request(before_request)
    # Allows for larger data to be sent through because of chunking types.
    app.before_request(handle_chunking)
    app.after_request(after_request)

    ServerEndpoints(app, tls_repo=tlsrepo, config=config)
    AdminEndpoints(app, tls_repo=tlsrepo, config=config)

    # TODO investigate socket-io connection here.
    # app.config['SECRET_KEY'] = 'secret!'
    # socketio = SocketIO(app)

    # @socketio.on('my event')
    # def handle_my_custom_event(json):
    #     print('received json: ' + str(json))
    #     send(f"I received: {json}")

    # now we get into the regular Flask details, except we're passing in the peer certificate
    # as a variable to the template.
    @app.route('/')
    def root():
        return redirect("/admin/index.html")
        # cert = request.environ['peercert']
        # cert_data = f"{cert.get_subject()}"
        # return render_template("admin/index.html")
        # return render_template('helloworld.html', client_cert=request.environ['peercert'])

    @app.route("/admin/index.html")
    def admin_home():
        return render_template("admin/index.html")

    @app.route("/admin/execute", methods=["get", "post"])
    def execute_method():
        if request.method == "POST":
            _log.debug("Posting stuff to server.")

        return render_template("admin/execute.html")

    @app.route("/admin/add-fsa", methods=["get", "post"])
    def admin_fsa():
        if request.method == "POST":
            return redirect("admin/index.html")

        controls, default_control = adpt.DERControlAdapter.get_all()
        return render_template("admin/add-fsa.html")

    @app.route("/admin/add-end-device", methods=["get", "post"])
    def admin_end_device():
        if request.method == "POST":
            return redirect(url_for("admin_home"))

        return render_template("admin/add-end-device.html", device_categories=DeviceCategoryType)

    @app.route("/admin/add-der-program", methods=["get", "post"])
    def admin_der_program():
        controls, default_control = adpt.DERControlAdapter.get_all()

        if request.method == "POST":
            args = request.form.to_dict()
            print(f"Args before: {args}")
            adpt.DERProgramAdapter.build(**args)
            print(f"Args after: {args}")

            return redirect(url_for("admin_home"))

        return render_template("admin/add-der-program.html",
                               der_controls=controls,
                               default_der_control=default_control)

    @app.route("/admin/default-der-control", methods=['get', 'post'])
    def admin_default_der_control():
        dderc = adpt.DERControlAdapter.fetch_default()

        if request.method == 'POST':
            kwargs = request.form.to_dict()
            # TODO: Build a helper that will allow us to populate by known form elements.
            # Helper for connect and energize mode, which are ready for usage.
            if 'enable_opModConnect' not in kwargs:
                kwargs['enable_opModConnect'] = 'off'

            if 'enable_opModEnergize' not in kwargs:
                kwargs['enable_opModEnergize'] = 'off'

            field_list = fields(dderc)
            base_control = dderc.DERControlBase
            for k, v in kwargs.items():
                if k.startswith('enable_'):
                    k = k.split('_')[1]
                    v = True if v == 'on' else False

                for f in field_list:
                    if k == f.name:
                        setattr(dderc, k, v)
                for f in fields(base_control):
                    if k == f.name:
                        setattr(base_control, k, v)

            adpt.DERControlAdapter.store_default(dderc=dderc)

            return redirect(url_for("admin_home"))

        return render_template("admin/update-default-der-control.html", dderc=dderc)

    @app.route("/admin/resources")
    def admin_resource_list():
        resource = request.args.get("rurl")
        obj = get_href(resource)
        all_resources = sorted(get_href_all_names())
        if obj:
            return render_template("admin/resource_list.html",
                                   resource_urls=all_resources,
                                   href_shown=resource,
                                   object=dataclass_to_xml(obj))
        else:
            return render_template("admin/resource_list.html", resource_urls=all_resources)

    @app.route("/admin/clients")
    def admin_clients():
        clients = tlsrepo.client_list
        return render_template("admin/clients.html", registered=clients, connected=[])

    @app.route("/admin/clients/dcap/<int:index>")
    def admin_clients_dcap(index: int = hrefs.NO_INDEX):
        clients = tlsrepo.client_list
        return render_template("admin/clients.html", registered=clients, connected=[])

    @app.route("/admin/routes")
    def admin_routes():
        routes = '<ul>'
        for p in app.url_map.iter_rules():
            routes += f"<li>{p.rule}</li>"
        routes += "</ul>"
        return Response(f"{routes}")

    return app


def run_app(app: Flask, host, ssl_context, request_handler, port, **kwargs):
    exclude_patterns = ["data_store/**", "docs/**", "examples/**", "ieee_2030_5_gui/**", "logs/**"]
    app.run(host=host,
            ssl_context=ssl_context,
            request_handler=request_handler,
            port=port,
            exclude_patterns=exclude_patterns,
            **kwargs)


def run_server(config: ServerConfiguration, tlsrepo: TLSRepository, **kwargs):
    global server_config, tls_repository
    server_config = config
    tls_repository = tlsrepo

    app = __build_app__(config, tlsrepo)

    ssl_context = None
    # If lfd_client is specified then we are running in http mode so we don'
    # establish an sslcontext.
    if not config.lfdi_client:
        ssl_context = __build_ssl_context__(tlsrepo)

    try:
        host, port = config.server_hostname.split(":")
    except ValueError:
        # host and port not available
        host = config.server_hostname
        port = 8443

    PeerCertWSGIRequestHandler.config = config
    PeerCertWSGIRequestHandler.tlsrepo = tlsrepo

    run_app(app=app,
            host=host,
            ssl_context=ssl_context,
            port=port,
            request_handler=PeerCertWSGIRequestHandler,
            **kwargs)


def build_server(config: ServerConfiguration, tlsrepo: TLSRepository, **kwargs) -> BaseWSGIServer:

    app = __build_app__(config, tlsrepo)
    ssl_context = __build_ssl_context__(tlsrepo)

    try:
        host, port = config.server_hostname.split(":")
    except ValueError:
        # host and port not available
        host = config.server_hostname
        port = 8443

    PeerCertWSGIRequestHandler.config = config
    PeerCertWSGIRequestHandler.tlsrepo = tlsrepo

    return make_server(app=app,
                       host=host,
                       ssl_context=ssl_context,
                       request_handler=PeerCertWSGIRequestHandler,
                       port=port,
                       **kwargs)

def make_app(config_file: Path, reset_certs: bool) -> Flask:
    config = ServerConfiguration.load(config_file)
    tlsrepo = TLSRepository(config.tls_repository, clear=reset_certs, openssl_cnffile_template=config.openssl_cnf, serverhost=config.server_hostname)
    app = __build_app__(config, tlsrepo)
    return app



from attrs import define, field
from gridappsd import GridAPPSD
from ieee_2030_5.certs import TLSRepository


@define
class GridAPPSDAdapter:
    gapps: GridAPPSD
    model_name: str
    model_id: str = field(default=None)

    def create_der_client_certificates(self, tls: TLSRepository):
        models = self.gapps.query_model_info()
        for m in models['data']['models']:
            if m['modelName'] == self.model_name:
                self.model_id = m['modelId']
                break
        if not self.model_id:
            raise ValueError(f"Model {self.model_name} not found")

        from cimgraph.data_profile import CIM_PROFILE
        import cimgraph.data_profile.rc4_2021 as cim
        from cimgraph.models import FeederModel
        from cimgraph.databases.gridappsd import GridappsdConnection
        from cimgraph.databases import ConnectionParameters

        cim_profile = CIM_PROFILE.RC4_2021.value
        iec = 7
        params = ConnectionParameters(cim_profile=cim_profile, iec61970_301=iec)

        conn = GridappsdConnection(params)
        conn.cim_profile = cim_profile
        feeder = cim.Feeder(mRID=self.model_id)

        network = FeederModel(connection=conn, container=feeder, distributed=False)

        network.get_all_edges(cim.PowerElectronicsConnection)

        for inv in network.graph[cim.PowerElectronicsConnection].values():
            tls.create_cert(inv.mRID)



from __future__ import annotations

import json
import logging
from datetime import datetime, timedelta
from http.client import BAD_REQUEST
from typing import Optional

import pytz
import tzlocal
import werkzeug.exceptions
from flask import Flask, Response, request
from werkzeug.exceptions import Forbidden
from werkzeug.routing import BaseConverter

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.data.indexer import get_href, get_href_filtered
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.server.dcapfs import Dcap
from ieee_2030_5.server.derfs import DERProgramRequests, DERRequests
from ieee_2030_5.server.enddevicesfs import (EDevRequests, FSARequests, SDevRequests)
# module level instance of hrefs class.
from ieee_2030_5.server.meteringfs import (MirrorUsagePointRequest, UsagePointRequest)
from ieee_2030_5.server.timefs import TimeRequest
from ieee_2030_5.server.uuid_handler import UUIDHandler
from ieee_2030_5.types_ import TimeOffsetType, format_time
from ieee_2030_5.utils import dataclass_to_xml, xml_to_dataclass

_log = logging.getLogger(__name__)


class Admin(RequestOp):

    def get(self):
        if not self.is_admin_client:
            raise Forbidden()
        return Response("We are able to do stuff here")

    def post(self):
        if not self.is_admin_client:
            raise Forbidden()
        return Response(json.dumps({'abc': 'def'}), headers={'Content-Type': 'application/json'})


class ServerList(RequestOp):

    def __init__(self, list_type: str, **kwargs):
        super().__init__(**kwargs)
        self._list_type = list_type

    def get(self) -> Response:
        response = None
        if self._list_type == 'EndDevice':
            response = self._end_devices.get_end_device_list(self.lfdi)

        if response:
            response = dataclass_to_xml(response)

        return response


class RegexConverter(BaseConverter):

    def __init__(self, url_map, *items):
        super(RegexConverter, self).__init__(url_map)
        self.regex = items[0]
        _log.debug(f"regex is {self.regex}")


class ServerEndpoints:

    def __init__(self, app: Flask, tls_repo: TLSRepository, config: ServerConfiguration):

        self.config = config
        self.tls_repo = tls_repo
        self.mimetype = "text/xml"
        self.app: Flask = app
        self.app.url_map.converters['regex'] = RegexConverter

        _log.debug(f"Adding rule: {hrefs.uuid_gen} methods: {['GET']}")
        app.add_url_rule(hrefs.uuid_gen, view_func=self._generate_uuid)
        _log.debug(f"Adding rule: {hrefs.get_dcap_href()} methods: {['GET']}")
        app.add_url_rule(hrefs.get_dcap_href(), view_func=self._dcap)
        _log.debug(f"Adding rule: {hrefs.get_time_href()} methods: {['GET']}")
        app.add_url_rule(hrefs.get_time_href(), view_func=self._tm)
        _log.debug(f"Adding rule: {hrefs.sdev} methods: {['GET']}")
        app.add_url_rule(hrefs.sdev, view_func=self._sdev)
        # _log.debug(f"Adding rule: {hrefs.derp} methods: {['GET']}")
        # app.add_url_rule(hrefs.derp, view_func=self._derp)

        # All the energy devices
        #app.add_url_rule(f"/{hrefs.EDEV}", methods=["GET", "POST", "PUT"], view_func=self._edev)
        app.add_url_rule(f"/<regex('{hrefs.EDEV}{hrefs.MATCH_REG}'):path>",
                         view_func=self._edev,
                         methods=["GET", "PUT", "POST"])
        # This rule must be before der
        app.add_url_rule(f"/<regex('{hrefs.DER_PROGRAM}{hrefs.MATCH_REG}'):path>",
                         view_func=self._derp,
                         methods=["GET"])
        app.add_url_rule(f"/<regex('{hrefs.DER}{hrefs.MATCH_REG}'):path>",
                         view_func=self._der,
                         methods=["GET", "PUT"])
        app.add_url_rule(f"/<regex('{hrefs.MUP}{hrefs.MATCH_REG}'):path>",
                         view_func=self._mup,
                         methods=["GET", "POST"])
        app.add_url_rule(f"/<regex('{hrefs.UTP}{hrefs.MATCH_REG}'):path>",
                         view_func=self._upt,
                         methods=["GET", "POST"])
        app.add_url_rule(f"/<regex('{hrefs.CURVE}{hrefs.MATCH_REG}'):path>",
                         view_func=self._curves,
                         methods=["GET"])

        app.add_url_rule(f"/<regex('{hrefs.FSA}{hrefs.MATCH_REG}'):path>",
                         view_func=self._fsa,
                         methods=["GET"])
        app.add_url_rule(f"/<regex('{hrefs.LOG}{hrefs.MATCH_REG}'):path>",
                         view_func=self._log,
                         methods=["GET", "POST"])
        # rulers = (
        #     (hrefs.der_urls, self._der),
        #     #(hrefs.edev_urls, self._edev),
        #     (hrefs.mup_urls, self._mup),
        #     (hrefs.curve_urls, self._curves),
        #     (hrefs.program_urls, self._programs)
        # )
        #
        # for endpoints, view_func in rulers:
        #     # Item should either be a single rule or a rule with a second element having the methods
        #     # in it.
        #     # edev = [
        #     #   /edev,
        #     #   (f"/edev/<int: index>", ["GET", "POST"])
        #     # ]
        #     for item in endpoints:
        #         try:
        #             rule, methods = item
        #         except ValueError:
        #             rule = item
        #             methods = ["GET"]
        #         _log.debug(f"Adding rule: {rule} methods: {methods}")
        #         app.add_url_rule(rule, view_func=view_func, methods=methods)
        #
        # self.add_endpoint(hrefs.dcap, view_func=self._dcap)
        # self.add_endpoint(hrefs.edev, view_func=self._edev)
        # self.add_endpoint(hrefs.mup, view_func=self._mup, methods=['GET', 'POST'])
        # self.add_endpoint(hrefs.uuid_gen, view_func=self._generate_uuid)
        # app.add_url_rule(hrefs.rsps, view_func=None)
        # self.add_endpoint(hrefs.tm, view_func=self._tm)
        #
        # for index, ed in end_devices.all_end_devices.items():
        #     self.add_endpoint(hrefs.edev + f"/{index}", view_func=self._edev)
        #     self.add_endpoint(hrefs.mup + f"/{index}", view_func=self._mup)

    def _log(self, path):
        return

    def _foo(self, bar):
        return Response("Foo Response")

    def _generate_uuid(self) -> Response:
        return Response(UUIDHandler().generate())

    #
    # def _admin(self) -> Response:
    #     return Admin(server_endpoints=self).execute()

    def _fsa(self, path) -> Response:
        return FSARequests(server_endpoints=self).execute()

    def _upt(self, path) -> Response:
        return UsagePointRequest(server_endpoints=self).execute()

    def _mup(self, path) -> Response:
        return MirrorUsagePointRequest(server_endpoints=self).execute()

    # Needs to be before der
    def _derp(self, path) -> Response:
        return DERProgramRequests(server_endpoints=self).execute()

    def _der(self, path) -> Response:
        _log.debug(request.method)
        return DERRequests(server_endpoints=self).execute()

    def _dcap(self) -> Response:
        return Dcap(server_endpoints=self).execute()

    def _edev(self, path: Optional[str] = None) -> Response:
        return EDevRequests(server_endpoints=self).execute()

    # def _edev(self, index: Optional[int] = None, category: Optional[str] = None) -> Response:
    #     return EDevRequests(server_endpoints=self).execute(index=index, category=category)

    def _sdev(self) -> Response:
        return SDevRequests(server_endpoints=self).execute()

    def _tm(self) -> Response:
        return TimeRequest(server_endpoints=self).execute()

    def _curves(self, path) -> Response:
        _list = adpt.ListAdapter.get_list(request.path)
        obj = m.DERCurveList(href=path, DERCurve=_list, all=len(_list))
        return RequestOp(server_endpoints=self).build_response_from_dataclass(obj)



from dataclasses import dataclass, field
import threading
from typing import Callable, Dict, List, Union
from ieee_2030_5.data.indexer import add_href
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
import ieee_2030_5.adapters as adpt
from ieee_2030_5.adapters import Adapter, NotFoundError, ResourceListAdapter
from ieee_2030_5.config import ReturnValue

from copy import deepcopy
from datetime import datetime
import time

from blinker import Signal
import logging

_log = logging.getLogger(__name__)

DERCurveAdapter = Adapter[m.DERCurve](hrefs.curve_href(), generic_type=m.DERCurve)
DERControlAdapter = Adapter[m.DERControl]("/derc", generic_type=m.DERControl)
DERCurveAdapter = Adapter[m.DERCurve](hrefs.curve_href(), generic_type=m.DERCurve)
DERProgramAdapter = Adapter[m.DERProgram](hrefs.der_program_href(), generic_type=m.DERProgram)
FunctionSetAssignmentsAdapter = Adapter[m.FunctionSetAssignments](
    url_prefix="/fsa", generic_type=m.FunctionSetAssignments)

EndDeviceAdapter = Adapter[m.EndDevice](hrefs.get_enddevice_href(), generic_type=m.EndDevice)
DeviceCapabilityAdapter = Adapter[m.DeviceCapability]("/dcap", generic_type=m.DeviceCapability)
# Generally the href will only be in the context of an end device.
RegistrationAdapter = Adapter[m.Registration](url_prefix="/reg", generic_type=m.Registration)
DERAdapter = Adapter[m.DER](url_prefix="/der", generic_type=m.DER)
UsagePointAdapter = Adapter[m.UsagePoint](url_prefix="/upt", generic_type=m.UsagePoint)
ListAdapter = ResourceListAdapter()


def _create_or_update_reading(reading_list_href: str, reading: m.Reading) -> ReturnValue:
    updated = False

    saved_reading: m.Reading | None = None
    # Attempt to find an existing reading with the same loc
    try:
        saved_reading = adpt.ListAdapter.get_item_by_prop(reading_list_href, 'localID',
                                                          reading.localID)
    except NotFoundError:
        # If not found then we add the reading to the list of readings.
        pass

    if saved_reading is None:
        saved_reading = adpt.ListAdapter.append_and_increment_href(reading_list_href, reading)
    else:
        updated = True
        adpt.ListAdapter.set(reading_list_href, int(saved_reading.href.split(hrefs.SEP)[-1]),
                             reading)

    return ReturnValue(True,
                       an_object=saved_reading,
                       was_update=updated,
                       location=saved_reading.href)


def _create_or_update_reading_set(reading_set_list_href: str,
                                  mirror_reading_set: m.MirrorReadingSet) -> ReturnValue:

    update = False
    current_mmr_reading_set: m.MirrorReadingSet | None = None

    current_mmr_reading_set = adpt.ListAdapter.get_by_mrid(mirror_reading_set.href,
                                                           mirror_reading_set.mRID)

    if current_mmr_reading_set is None:
        mmr_href = reading_set_list_href.replace(hrefs.DEFAULT_UPT_ROOT, hrefs.DEFAULT_MUP_ROOT)
        current_mmr_reading_set = adpt.ListAdapter.append_and_increment_href(
            mmr_href, mirror_reading_set)
        new_rs = m.ReadingSet(description=mirror_reading_set.description,
                              timePeriod=mirror_reading_set.timePeriod,
                              version=mirror_reading_set.version)
        current_rs = adpt.ListAdapter.append_and_increment_href(reading_set_list_href, new_rs)
    else:
        update = True
        current_rs = adpt.ListAdapter.get(reading_set_list_href,
                                          key=int(mirror_reading_set.href.split(hrefs.SEP)[-1]))
        current_rs.description = mirror_reading_set.description
        current_rs.timePeriod = mirror_reading_set.timePeriod
        current_rs.version = mirror_reading_set.version
        adpt.ListAdapter.set(reading_set_list_href,
                             int(mirror_reading_set.href.split(hrefs.SEP)[-1]), current_rs)

    return ReturnValue(True, current_rs, update, current_rs.href)

    # # Update the href of the mirror reading set. so it can be referenced correctly
    # # within the list of mirror reading sets.
    # new_rs_index = adpt.ListAdapter.list_size(reading_set_list_href)
    # mirror_rs_href = hrefs.SEP.join([mirror_reading_set.href, str(new_rs_index)])
    # mirror_reading_set.href = mirror_rs_href

    # adpt.ListAdapter.append(mirror_reading_set.href), mirror_reading_set)

    # # new_rs: m.ReadingSet = m.ReadingSet(href=hrefs.SEP.join([reading_set_list_href, str(new_rs_index)]),
    # #                       description=mirror_reading_set.description,
    # #                       timePeriod=mirror_reading_set.timePeriod,
    # #                       version=mirror_reading_set.version,
    # #                       ReadingTypeLink=adpt.ListAdapter.get_single(mirror_reading_set.ReadingTypeLink,)
    # # # saved_rs: m.ReadingSet = adpt.ListAdapter.append_and_increment_href(reading_set_list_href, new_rs)

    # # # adpt.ListAdapter.append(reading_set_list_href),
    # # #                         mirror_reading_set.Reading)


def _create_or_update_meter_reading(
    upt_list_href: str,
    mmr: m.MirrorMeterReading,
) -> ReturnValue:
    """
    Create a MeterReading from a MirrorMeterReading and add it to the UsagePoint.

    Note: The passed mmr MUST have the uri set ot the actual mmr not the list!

    """

    saved_mr = adpt.ListAdapter.get_by_mrid(upt_list_href, mmr.mRID)

    was_updated = False
    if not saved_mr:
        # Create the new meter reading from the mirror meter reading.
        new_mr = m.MeterReading(mRID=mmr.mRID, description=mmr.description)

        saved_mr: m.MeterReading = adpt.ListAdapter.append_and_increment_href(
            upt_list_href, new_mr)
    else:
        was_updated = True
        saved_mr.description = mmr.description

    assert saved_mr.href.split(hrefs.SEP)[-1] == mmr.href.split(hrefs.SEP)[-1]

    return ReturnValue(True, was_update=was_updated, an_object=saved_mr, location=saved_mr.href)

    # mmr.ReadingSet is a list of MirrorReadingSet objects.
    if mmr.MirrorReadingSet is not None:
        meter_reading_set_list_href = hrefs.SEP.join([saved_mr.href, "rs"])
        for mmr_rs in mmr.MirrorReadingSet:
            stored_mmrs: m.MirrorReadingSet = adpt.ListAdapter.append_and_increment_href(
                hrefs.SEP.join([mmr.href, "rs"]), mmr_rs)
            ret = _create_or_update_reading_set(meter_reading_set_list_href, stored_mmrs)
            assert isinstance(ret.an_object, m.ReadingSet)
            myrs: m.ReadingSet = ret.an_object

            for mmr_r in mmr_rs.Reading:
                new_mmr_r_list_href = hrefs.SEP.join([stored_mmrs.href, "r"])
                saved_mmr_reading = adpt.ListAdapter.append_and_increment_href(
                    new_mmr_r_list_href, mmr_r)

                new_r_list_href = hrefs.SEP.join([myrs.href, "r"])
                copy_of_reading = deepcopy(mmr_r)
                saved_reading = adpt.ListAdapter.append_and_increment_href(
                    new_r_list_href, copy_of_reading)
                assert saved_mmr_reading.href.split(hrefs.SEP)[-1] == saved_reading.href.split(
                    hrefs.SEP)[-1]

        upt_mrs_href = hrefs.SEP.join([saved_mr.href, "rs"])

        ret: ReturnValue = _create_or_update_reading_set(upt_mrs_href, stored_mmrs)
        new_rs: m.ReadingSet = ret.an_object
        saved_mr.ReadingSetListLink = m.ReadingSetListLink(href=upt_mrs_href)
        adpt.ListAdapter.set(upt_list_href, int(saved_mr.href.split(hrefs.SEP)[-1]), saved_mr)

        # rs_list: List[m.ReadingSet] | None = None
        # try:
        #     rs_list = adpt.ListAdapter.get_list(upt_mrs_href)
        # except KeyError:
        #     adpt.ListAdapter.initialize_uri(upt_mrs_href, m.ReadingSet)

        # if rs_list is None:
        #     new_rs_index = 0
        # else:
        #     new_rs_index = len(rs_list)

        # new_rs_href = hrefs.SEP.join([upt_mrs_href, str(new_rs_index)])
        # new_rs = m.ReadingSet(href=new_rs_href, description=mmr.MirrorReadingSet.description,
        #                       timePeriod=mmr.MirrorReadingSet.timePeriod, version=mmr.MirrorReadingSet.version,
        #                       ReadingListLink=m.ReadingListLink(href=hrefs.SEP.join([new_rs_href, "r"])))

        #rs = m.ReadingSet(href=upt_mrs_href, description=mmr.MirrorReadingSet.description)
        # adpt.ListAdapter.append(upt_mrs_href, ).initialize_uri(upt_mrs_href, m.ReadingSet)

        # new_mr.ReadingSetListLink = m.ReadingSetListLink(href=upt_mrs_href)

    # Create a meter reading for the usage point.
    # upt_meter_reading = m.MeterReading(href=new_mmr_href,
    #                                    mRID=mmr.mRID,
    #                                    description=mmr.description)
    # adpt.ListAdapter.append(upt.MeterReadingListLink.href, upt_meter_reading)

    # # Store the type of reading using the add_href method.
    # rt_href = hrefs.SEP.join([upt.href, "rt", str(new_mmr_index)])
    # upt_meter_reading.ReadingTypeLink = m.ReadingTypeLink(rt_href)
    # add_href(rt_href, mmr.ReadingType)

    return saved_mr


def create_or_update_meter_reading(
    mup_href: str,
    mmr_input: Union[m.MirrorMeterReading, m.MirrorMeterReadingList],
) -> ReturnValue:

    if isinstance(mmr_input, m.MirrorMeterReadingList):
        raise NotImplemented()

    mup: m.MirrorUsagePoint = adpt.ListAdapter.get(hrefs.DEFAULT_MUP_ROOT,
                                                   mup_href.split(hrefs.SEP)[-1])
    assert isinstance(mup, m.MirrorUsagePoint)

    upt: m.UsagePoint = adpt.ListAdapter.get(hrefs.DEFAULT_UPT_ROOT, mup_href.split(hrefs.SEP)[-1])

    assert isinstance(upt, m.UsagePoint)

    was_updated = False
    location = None

    if isinstance(mmr_input, m.MirrorMeterReadingList):
        return ReturnValue(False, "Not Implemented")

    elif isinstance(mmr_input, m.MirrorMeterReading):

        # Attempt to find an existing mirror meter reading with the same mrid.
        # If found then we need to replace it with this new meter reading.
        mmr_list_href = hrefs.SEP.join([mup_href, "mr"])
        upt_list_href = mmr_list_href.replace("mup", "upt")

        mmr_current: m.MirrorMeterReading = adpt.ListAdapter.get_by_mrid(
            mmr_list_href, mmr_input.mRID)

        # Each of the following will only create a single level and then return.  The return
        # object will be the last object created.
        #
        # if not found then we add the mmr to the list of mmrs.
        if mmr_current is None:
            # Store the new mirror meter reading.
            saved_mmr = adpt.ListAdapter.append_and_increment_href(mmr_list_href, mmr_input)
        else:
            saved_mmr = mmr_current

        # Create a meter reading for the usage point from the mirror meter reading.
        meter_reading_result = _create_or_update_meter_reading(upt_list_href, saved_mmr)
        meter_reading = meter_reading_result.an_object

        if mmr_input.ReadingType is not None:
            rt_href = hrefs.SEP.join([meter_reading.href, "rt"])
            adpt.ListAdapter.set_single(rt_href, mmr_input.ReadingType)
            meter_reading.ReadingTypeLink = m.ReadingTypeLink(href=rt_href)

        if mmr_input.Reading is not None:
            upt_r_href = hrefs.SEP.join([meter_reading.href, "r"])
            meter_reading.ReadingListLink = m.ReadingListLink(href=upt_r_href)
            reading_result = _create_or_update_reading(upt_r_href, mmr_input.Reading)

        # Use the meter_reading_results object
        if mmr_input.MirrorReadingSet is not None:
            upt_rs_href = hrefs.SEP.join([meter_reading_result.an_object.href, "rs"])
            meter_reading.ReadingSetListLink = m.ReadingSetListLink(href=upt_rs_href)

            for mmr_rs in mmr_input.MirrorReadingSet:
                ret = _create_or_update_reading_set(upt_rs_href, mmr_rs)
                assert isinstance(ret.an_object, m.ReadingSet)
                myrs: m.ReadingSet = ret.an_object

                upt_r_href = hrefs.SEP.join([ret.an_object.href, "r"])
                myrs.ReadingListLink = m.ReadingListLink(href=upt_r_href)
                for mmr_r in mmr_rs.Reading:
                    reading_result = _create_or_update_reading(upt_r_href, mmr_r)

        adpt.ListAdapter.store()
        # The over all result of this call!
        return meter_reading_result

        # new_mmr_index = adpt.ListAdapter.list_size(mmr_list_href)
        # new_mmr_href = hrefs.SEP.join([mmr_list_href, str(new_mmr_index)])
        # mmr_input.href = new_mmr_href

        # upt_meter_reading = _create_or_update_meter_reading(upt, mmr_input)

        # if mmr_input.MirrorReadingSet is not None:
        #     return_value = _create_or_update_reading_set(mup, mmr_input.MirrorReadingSet, internal_ref_transform)

    #     # Attempt to find an existing mup with the same mrid.  If found then we need to replace
    #     # the data with the new data etc.  If not found then we add the mmr to the list of mmrs.
    #     mmr_list_href = hrefs.SEP.join([mup_href, "mr"])
    #     upt_list_href = mmr_list_href.replace("mup", "upt")
    #     mmr_item = adpt.ListAdapter.get_by_mrid(mmr_list_href), mmr_input.mRID)

    #     if mmr_item:
    #         mmr_index = adpt.ListAdapter.get_list(mmr_list_href)).index(mmr_item)
    #         was_updated = True
    #     else:
    #         # This is a new mirror meter reading index.
    #         new_mmr_index = adpt.ListAdapter.list_size(mmr_list_href))

    #         # Initialize the url for the mirror meter reading list.
    #         mmr_input.href = hrefs.SEP.join([mmr_list_href, str(new_mmr_index)])

    #         adpt.ListAdapter.initialize_uri(upt_list_href), m.ReadingList)
    #         # Add a new mirror meter reading to the list of mirror meter readings.
    #         mmr_input.href = hrefs.SEP.join([mmr_list_href, str(new_mmr_index)])

    #         # If this is a List of MirrorMeterReadings then we need to add each one to the list. And
    #         # initialize the types of readings.
    #         if mup.MirrorMeterReading:
    #             for mup_mr in mup.MirrorMeterReading:
    #                 mmr_input.ReadingType = mup_mr.ReadingType
    #                 mmr_input.description = mup_mr.description
    #                 mmr_input.lastUpdateTime = mup_mr.lastUpdateTime

    #             ListAdapter.append(mmr_list_href), mmr_input)
    #             location = mmr_input.href

    #         if mmr_input.Reading is not None and mmr_input.ReadingType is not None:
    #             stored_mirror_reading = adpt.ListAdapter.get_by_mrid(mmr_list_href), mmr_input.mRID)

    #             # If we found it, then replace with new reading
    #             if stored_mirror_reading:
    #                 mirror_reading_indx = adpt.ListAdapter.get_list(mmr_list_href)).index(stored_mirror_reading)

    #                 # Replace MirrorMeterReading with new data
    #                 adpt.ListAdapter.set(mmr_list_href), mirror_reading_indx, mmr_input)

    #                 adpt.ListAdapter.set(upt_list_href), mirror_reading_indx, mmr_input.Reading)
    #                 was_updated = True
    #             else:
    #                 reading_index = adpt.ListAdapter.list_size(mmr_list_href))
    #                 mmr_input.Reading.href = hrefs.SEP.join([mmr_input.href, "r"])
    #                 ListAdapter.append(mmr_list_href), mmr_input)
    #                 ListAdapter.set(upt_list_href), reading_index, mmr_input)
    #                 location = mmr_input.href

    #             # ListAdapter.set_single(mmr_list_href), mmr_input)
    #             # upt_input = deepcopy(mmr_input)
    #             # upt_input.href = upt_input.href.replace("mup", "upt")
    #             # ListAdapter.set_single(upt_list_href), upt_input)
    #             # location = upt_list_href

    #     return ReturnValue(True, mmr_input, was_updated, location)

    #     try:
    #         # Check to see if the meter reading already exists based upon the passed mRID.
    #         mmr_item = adpt.ListAdapter.get_by_mrid(mmr_list_href), mmr_input.mRID)
    #         new_mmr_index = adpt.ListAdapter.get_list(mmr_list_href)).index(mmr_item)
    #         was_updated = True
    #     except NotFoundError:
    #         new_mmr_index = adpt.ListAdapter.list_size(mmr_list_href))
    #         mmr_input.href = hrefs.SEP.join([mmr_list_href, str(new_mmr_index)])

    #         # This is a singleton reading here.
    #         if mmr_input.ReadingType is not None and mmr_input.Reading is not None:
    #             ListAdapter.set_single(mmr_list_href), mmr_input)

    #         for mup_mr in mup.MirrorMeterReading:
    #             mmr_input.ReadingType = mup_mr.ReadingType
    #             mmr_input.description = mup_mr.description
    #             mmr_input.lastUpdateTime = mup_mr.lastUpdateTime
    #         ListAdapter.append(mmr_list_href), mmr_input)
    #         mmr_item = mmr_input

    #     try:
    #         mr = adpt.ListAdapter.get(upt_list_href), new_mmr_index)
    #     except NotFoundError:
    #         # This shouldn't happen if it does then there is something wrong with our code.
    #         if was_updated:
    #             raise ValueError("Unable to find meter reading for updated mirror meter reading")
    #         rt_href = hrefs.SEP.join([upt_list_href, str(new_mmr_index), "rt"])
    #         mmr_item.ReadingType.href = rt_href
    #         mr = m.MeterReading(href=hrefs.SEP.join([upt_list_href, str(new_mmr_index)]),
    #                             mRID=mmr_item.mRID,
    #                             description=mmr_item.description,
    #                             ReadingTypeLink=m.ReadingTypeLink(rt_href))
    #         add_href(rt_href, mmr_item.ReadingType)
    #         ListAdapter.append(upt_list_href), mr)
    #     location = upt_list_href

    #     # Current instantanious values.
    #     if mmr_input.Reading is not None:
    #         mmr_item.Reading.href = hrefs.SEP.join([mmr_item.href, "r"])
    #         mmr_item.Reading.href = mmr_item.Reading.href.replace("mup", "upt")
    #         add_href(mmr_item.Reading.href, mmr_item.Reading)
    #         mr.ReadingLink = m.ReadingLink(mmr_item.Reading.href)

    #     # Mirror reading sets
    #     if mmr_input.MirrorReadingSet:

    #         mrs_list_href = hrefs.SEP.join([mmr_item.href, "rs"])
    #         rs_list_href = hrefs.SEP.join([mmr_item.href, "rs"]).replace("mup", "upt")
    #         mr.ReadingSetListLink = m.ReadingSetListLink(href=rs_list_href)
    #         ListAdapter.initialize_uri(mr.ReadingSetListLink.href), m.ReadingSet)

    #         for mrs in mmr_input.MirrorReadingSet:
    #             found_rs = False
    #             try:
    #                 mrs_item = ListAdapter.get_by_mrid(mrs_list_href), mrs.mRID)
    #                 mrs_item_index = ListAdapter.get_list(mrs_list_href).index(mrs_item)
    #                 found_rs = True
    #             except NotFoundError:
    #                 mrs_item = mrs
    #                 mrs_item_index = ListAdapter.list_size(mrs_list_href))
    #                 mrs_item.href = hrefs.SEP.join([mrs_list_href, str(mrs_item_index)])
    #                 ListAdapter.append(mrs_list_href), mrs_item)

    #             if found_rs:
    #                 rs_item = ListAdapter.get(rs_list_href), mrs_item_index)
    #                 rs_item.description = mrs_item.description
    #                 rs_item.timePeriod = mrs_item.timePeriod
    #                 rs_item.version = mrs_item.version
    #             else:
    #                 rs_item = m.ReadingSet(href=hrefs.SEP.join([rs_list_href,
    #                                                             str(mrs_item_index)]),
    #                                        description=mrs_item.description,
    #                                        timePeriod=mrs_item.timePeriod,
    #                                        version=mrs_item.version)
    #                 ListAdapter.append(rs_list_href), rs_item)

    #             reading_list_href = hrefs.SEP.join([rs_item.href, "r"])
    #             rs_item.ReadingListLink = m.ReadingListLink(href=reading_list_href)
    #             for reading_index, reading in enumerate(mrs_item.Reading):
    #                 reading.href = hrefs.SEP.join([reading_list_href, str(reading_index)])
    #                 ListAdapter.append(reading_list_href), reading)

    # ListAdapter.store()

    return ReturnValue(True, mmr_item, was_updated, location)


def create_mirror_usage_point(mup: m.MirrorUsagePoint, ) -> ReturnValue:
    """Creates a MirrorUsagePoint and associated UsagePoint and adds them to their adapters.
    """

    # Attempt to find an existing mup with the same mrid.  If found then we need to replace
    # it with the new data etc.
    found_with_mrid = None
    if adpt.ListAdapter.list_size(hrefs.DEFAULT_MUP_ROOT) > 0:
        try:
            found_with_mrid = adpt.ListAdapter.get_by_mrid(hrefs.DEFAULT_MUP_ROOT, mup.mRID)
        except NotFoundError:
            ...

    update = False
    if not found_with_mrid:
        # Creating a new allocation of resources for the mup. And then copy the data from the
        # mup resources into the usage point resources allocating new data as needed.
        upt_href = hrefs.UsagePointHref()

        # Both the usage point and mirror usage point will have the same point index.
        point_index = adpt.ListAdapter.list_size(hrefs.DEFAULT_MUP_ROOT)
        mup.href = hrefs.SEP.join([hrefs.DEFAULT_MUP_ROOT, str(point_index)])
        # Add the mirror usage point to the list of mirror usage points.
        adpt.ListAdapter.append(hrefs.DEFAULT_MUP_ROOT, mup)

        # Create a usage point with the same index as the mirror usage point.
        upt = m.UsagePoint(href=hrefs.SEP.join([hrefs.DEFAULT_UPT_ROOT,
                                                str(point_index)]),
                           description=mup.description,
                           deviceLFDI=mup.deviceLFDI,
                           serviceCategoryKind=mup.serviceCategoryKind,
                           mRID=mup.mRID,
                           roleFlags=mup.roleFlags,
                           status=mup.status)
        upt.MeterReadingListLink = m.MeterReadingListLink(href=hrefs.SEP.join([upt.href, "mr"]),
                                                          all=0)
        adpt.ListAdapter.append(hrefs.DEFAULT_UPT_ROOT, upt)

        # Initialize the url for the mirror meter reading list.
        mmr_list_href = hrefs.SEP.join([mup.href, "mr"])
        adpt.ListAdapter.initialize_uri(mmr_list_href, m.MirrorMeterReading)

        # Initialize the url for the meter reading list.
        mr_list_href = hrefs.SEP.join([upt.href, "mr"])
        adpt.ListAdapter.initialize_uri(mr_list_href, m.MeterReading)

        for index_for_readings, mirror_meter_reading in enumerate(mup.MirrorMeterReading):
            # Validate the the reading has a reading type associated with it.
            if not mirror_meter_reading.ReadingType:
                return ReturnValue(
                    False,
                    f"Invalid Reading Type for Mirror Meter Reading {mirror_meter_reading.mRID}")

            # Update the mirror meter reading href and then add it to the list of mirror meter readings.
            mirror_meter_reading.href = hrefs.SEP.join([mmr_list_href, str(index_for_readings)])
            adpt.ListAdapter.append(mmr_list_href, mirror_meter_reading)

            # Create a meter reading for the usage point.
            meter_reading = m.MeterReading(href=hrefs.SEP.join(
                [mr_list_href, str(index_for_readings)]),
                                           mRID=mirror_meter_reading.mRID,
                                           description=mirror_meter_reading.description)
            adpt.ListAdapter.append(mr_list_href, meter_reading)

            # Store the type of reading using the add_href method.
            rt_href = hrefs.SEP.join([upt.href, "rt", str(index_for_readings)])
            meter_reading.ReadingTypeLink = m.ReadingTypeLink(rt_href)

            adpt.ListAdapter.set_single(rt_href, mirror_meter_reading.ReadingType)
            add_href(rt_href, mirror_meter_reading.ReadingType)

    else:
        # TODO Update all properties with new items from mup
        mup.href = found_with_mrid.href
        found_with_mrid.description = mup.description
        found_with_mrid.deviceLFDI = mup.deviceLFDI
        found_with_mrid.serviceCategoryKind = mup.serviceCategoryKind
        found_with_mrid.mRID = mup.mRID
        found_with_mrid.roleFlags = mup.roleFlags
        found_with_mrid.status = mup.status
        update = True
        adpt.ListAdapter.store()

    return ReturnValue(True, mup, update, mup.href)


@dataclass
class TimerSpec:
    trigger_after_seconds: int
    fn: Callable
    args: List = field(default_factory=list)
    kwargs: Dict = field(default_factory=dict)
    enabled: bool = True
    trigger_count: int = 0
    last_trigger_time: int = int(time.mktime(datetime.utcnow().timetuple()))

    def disable(self):
        self.enabled = False

    def enable(self):
        self.enabled = True

    def reset_count(self):
        self.trigger_count = 0

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TimerSpec):
            raise NotImplementedError(
                f"Comparison between {self.__class__.__name__} and {type(other)} not implemented")
        return self.fn is other.fn

    def trigger(self, current_time: int):
        if self.last_trigger_time + self.trigger_after_seconds < current_time:
            if self.args and self.kwargs:
                self.fn(args=self.args, kwargs=self.kwargs)
            elif self.args:
                self.fn(args=self.args)
            else:
                self.fn()
            self.trigger_count += 1
            self.last_trigger_time = current_time


class _TimeAdapter(threading.Thread):
    tick = Signal("tick")
    event_started = Signal("event_started")
    event_ended = Signal("event_endend")
    event_scheduled = Signal("event_scheduled")
    events: Dict[str, m.Event] = {}
    current_tick: int = 0
    looping: bool = False

    @staticmethod
    def user_readable(timestamp: int) -> str:
        dt = datetime.fromtimestamp(timestamp)
        return dt.isoformat()    # .strftime("%m/%d/%Y, %H:%M:%S")

    @staticmethod
    def from_iso(iso_fmt_date: str) -> int:
        dt = datetime.strptime(iso_fmt_date, "%Y-%m-%dT%H:%M:%S")
        return int(time.mktime(dt.timetuple()))

    @staticmethod
    def add_event(evnt: m.Event):
        time_now = _TimeAdapter.current_tick
        if evnt.EventStatus is None:
            evnt.EventStatus = m.EventStatus()
        if evnt.href not in _TimeAdapter.events:
            while _TimeAdapter.looping:
                time.sleep(0.1)
            _TimeAdapter.events[evnt.href] = evnt

    def run(self) -> None:

        while True:
            _TimeAdapter.current_tick = int(time.mktime(datetime.utcnow().timetuple()))
            _TimeAdapter.tick.send(self.current_tick)
            _TimeAdapter.looping = True
            time_now = _TimeAdapter.current_tick
            for href, evnt in _TimeAdapter.events.items():
                if time_now < evnt.interval.start and evnt.EventStatus.currentStatus is None:
                    evnt.EventStatus.dateTime = time_now
                    evnt.EventStatus.currentStatus = 0
                    _log.debug(f"{'='*20}Event Scheduled {evnt.href}")
                    _TimeAdapter.event_scheduled.send(evnt)
                elif evnt.interval.start < time_now and time_now < evnt.interval.start + evnt.interval.duration:
                    if evnt.EventStatus.currentStatus != 1:
                        evnt.EventStatus.currentStatus = 1
                        evnt.EventStatus.dateTime = time_now
                        _log.debug(f"{'='*20}Event Started {evnt.href}")
                        _TimeAdapter.event_started.send(evnt)
                elif time_now > evnt.interval.start + evnt.interval.duration and evnt.EventStatus.currentStatus == 1:
                    evnt.EventStatus.currentStatus = 5
                    evnt.EventStatus.dateTime = time_now
                    _log.debug(f"{'='*20}Event Complete {evnt.href}")
                    _TimeAdapter.event_ended.send(evnt)

            _TimeAdapter.looping = False
            time.sleep(1)


TimeAdapter = _TimeAdapter()
TimeAdapter.daemon = True
TimeAdapter.start()



from ieee_2030_5.models.Config import (
    TypeName,
    CompoundFields,
    Config,
    ConstantName,
    Conventions,
    FieldName,
    Format,
    ModuleName,
    Output,
    PackageName,
    Substitution,
    Substitutions,
)
from ieee_2030_5.models.derforecasts import (DERFlexibility, DERForecast, DERForecastLink,
                                             ForecastNumericType, ForecastParameter,
                                             ForecastParameterSet, ForecastParameterSetList)
from ieee_2030_5.models.enums import CurveType, DeviceCategoryType, PrimacyType
from ieee_2030_5.models.sep import (
    DER, IEEE_802_15_4, AbstractDevice, AccountBalance, AccountBalanceLink, AccountingUnit,
    ActiveBillingPeriodListLink, ActiveCreditRegisterListLink, ActiveDERControlListLink,
    ActiveEndDeviceControlListLink, ActiveFlowReservationListLink, ActivePower,
    ActiveProjectionReadingListLink, ActiveSupplyInterruptionOverrideListLink,
    ActiveTargetReadingListLink, ActiveTextMessageListLink, ActiveTimeTariffIntervalListLink,
    AmpereHour, ApparentPower, ApplianceLoadReduction, AppliedTargetReduction,
    AssociatedDERProgramListLink, AssociatedUsagePointLink, BillingMeterReadingBase, BillingPeriod,
    BillingPeriodList, BillingPeriodListLink, BillingReading, BillingReadingList,
    BillingReadingListLink, BillingReadingSet, BillingReadingSetList, BillingReadingSetListLink,
    Charge, Condition, Configuration, ConfigurationLink, ConnectStatusType,
    ConsumptionTariffInterval, ConsumptionTariffIntervalList, ConsumptionTariffIntervalListLink,
    CreditRegister, CreditRegisterList, CreditRegisterListLink, CreditTypeChange,
    CurrentDERProgramLink, CurrentRMS, CurveData, CustomerAccount, CustomerAccountLink,
    CustomerAccountList, CustomerAccountListLink, CustomerAgreement, CustomerAgreementList,
    CustomerAgreementListLink, DateTimeInterval, DefaultDERControl, DefaultDERControlLink,
    DemandResponseProgram, DemandResponseProgramLink, DemandResponseProgramList,
    DemandResponseProgramListLink, DERAvailability, DERAvailabilityLink, DERCapability,
    DERCapabilityLink, DERControl, DERControlBase, DERControlList, DERControlListLink,
    DERControlResponse, DERCurve, DERCurveLink, DERCurveList, DERCurveListLink, DERLink, DERList,
    DERListLink, DERProgram, DERProgramLink, DERProgramList, DERProgramListLink, DERSettings,
    DERSettingsLink, DERStatus, DERStatusLink, DeviceCapability, DeviceCapabilityLink,
    DeviceInformation, DeviceInformationLink, DeviceStatus, DeviceStatusLink, DRLCCapabilities,
    DrResponse, DutyCycle, EndDevice, EndDeviceControl, EndDeviceControlList,
    EndDeviceControlListLink, EndDeviceLink, EndDeviceList, EndDeviceListLink, EnvironmentalCost,
    Error, Event, EventStatus, File, FileLink, FileList, FileListLink, FileStatus, FileStatusLink,
    FixedPointType, FixedVar, FlowReservationRequest, FlowReservationRequestList,
    FlowReservationRequestListLink, FlowReservationResponse, FlowReservationResponseList,
    FlowReservationResponseListLink, FlowReservationResponseResponse, FreqDroopType,
    FunctionSetAssignments, FunctionSetAssignmentsBase, FunctionSetAssignmentsList,
    FunctionSetAssignmentsListLink, GPSLocationType, HistoricalReading, HistoricalReadingList,
    HistoricalReadingListLink, IdentifiedObject, InverterStatusType, IPAddr, IPAddrList,
    IPAddrListLink, IPInterface, IPInterfaceList, IPInterfaceListLink, Link, List_type, ListLink,
    LLInterface, LLInterfaceList, LLInterfaceListLink, LoadShedAvailability,
    LoadShedAvailabilityList, LoadShedAvailabilityListLink, LocalControlModeStatusType, LogEvent,
    LogEventList, LogEventListLink, ManufacturerStatusType, MessagingProgram, MessagingProgramList,
    MessagingProgramListLink, MeterReading, MeterReadingBase, MeterReadingLink, MeterReadingList,
    MeterReadingListLink, MirrorMeterReading, MirrorMeterReadingList, MirrorReadingSet,
    MirrorUsagePoint, MirrorUsagePointList, MirrorUsagePointListLink, Neighbor, NeighborList,
    NeighborListLink, Notification, NotificationList, NotificationListLink, Offset,
    OperationalModeStatusType, PEVInfo, PowerConfiguration, PowerFactor, PowerFactorWithExcitation,
    PowerStatus, PowerStatusLink, Prepayment, PrepaymentLink, PrepaymentList, PrepaymentListLink,
    PrepayOperationStatus, PrepayOperationStatusLink, PriceResponse, PriceResponseCfg,
    PriceResponseCfgList, PriceResponseCfgListLink, ProjectionReading, ProjectionReadingList,
    ProjectionReadingListLink, RandomizableEvent, RateComponent, RateComponentLink,
    RateComponentList, RateComponentListLink, ReactivePower, ReactiveSusceptance, Reading,
    ReadingBase, ReadingLink, ReadingList, ReadingListLink, ReadingSet, ReadingSetBase,
    ReadingSetList, ReadingSetListLink, ReadingType, ReadingTypeLink, RealEnergy, Registration,
    RegistrationLink, RequestStatus, Resource, RespondableIdentifiedObject, RespondableResource,
    RespondableSubscribableIdentifiedObject, Response, ResponseList, ResponseListLink, ResponseSet,
    ResponseSetList, ResponseSetListLink, RPLInstance, RPLInstanceList, RPLInstanceListLink,
    RPLSourceRoutes, RPLSourceRoutesList, RPLSourceRoutesListLink, SelfDevice, SelfDeviceLink,
    ServiceChange, ServiceSupplier, ServiceSupplierLink, ServiceSupplierList, SetPoint,
    SignedRealEnergy, StateOfChargeStatusType, StorageModeStatusType, SubscribableIdentifiedObject,
    SubscribableList, SubscribableResource, Subscription, SubscriptionBase, SubscriptionList,
    SubscriptionListLink, SupplyInterruptionOverride, SupplyInterruptionOverrideList,
    SupplyInterruptionOverrideListLink, SupportedLocale, SupportedLocaleList,
    SupportedLocaleListLink, TargetReading, TargetReadingList, TargetReadingListLink,
    TargetReduction, TariffProfile, TariffProfileLink, TariffProfileList, TariffProfileListLink,
    Temperature, TextMessage, TextMessageList, TextMessageListLink, TextResponse, Time,
    TimeConfiguration, TimeLink, TimeTariffInterval, TimeTariffIntervalList,
    TimeTariffIntervalListLink, UnitValueType, UnsignedFixedPointType, UsagePoint, UsagePointBase,
    UsagePointLink, UsagePointList, UsagePointListLink, VoltageRMS, WattHour, loWPAN)

__all__ = [
    "TypeName",
    "CompoundFields",
    "Config",
    "ConstantName",
    "Conventions",
    "FieldName",
    "Format",
    "ModuleName",
    "Output",
    "PackageName",
    "Substitution",
    "Substitutions",
    "DERFlexibility",
    "DERForecast",
    "DERForecastLink",
    "ForecastNumericType",
    "ForecastParameter",
    "ForecastParameterSet",
    "ForecastParameterSetList",
    "AbstractDevice",
    "AccountBalance",
    "AccountBalanceLink",
    "AccountingUnit",
    "ActiveBillingPeriodListLink",
    "ActiveCreditRegisterListLink",
    "ActiveDERControlListLink",
    "ActiveEndDeviceControlListLink",
    "ActiveFlowReservationListLink",
    "ActivePower",
    "ActiveProjectionReadingListLink",
    "ActiveSupplyInterruptionOverrideListLink",
    "ActiveTargetReadingListLink",
    "ActiveTextMessageListLink",
    "ActiveTimeTariffIntervalListLink",
    "AmpereHour",
    "ApparentPower",
    "ApplianceLoadReduction",
    "AppliedTargetReduction",
    "AssociatedDERProgramListLink",
    "AssociatedUsagePointLink",
    "BillingMeterReadingBase",
    "BillingPeriod",
    "BillingPeriodList",
    "BillingPeriodListLink",
    "BillingReading",
    "BillingReadingList",
    "BillingReadingListLink",
    "BillingReadingSet",
    "BillingReadingSetList",
    "BillingReadingSetListLink",
    "Charge",
    "Condition",
    "Configuration",
    "ConfigurationLink",
    "ConnectStatusType",
    "ConsumptionTariffInterval",
    "ConsumptionTariffIntervalList",
    "ConsumptionTariffIntervalListLink",
    "CreditRegister",
    "CreditRegisterList",
    "CreditRegisterListLink",
    "CreditTypeChange",
    "CurrentDERProgramLink",
    "CurrentRMS",
    "CurveData",
    "CurveType",
    "CustomerAccount",
    "CustomerAccountLink",
    "CustomerAccountList",
    "CustomerAccountListLink",
    "CustomerAgreement",
    "CustomerAgreementList",
    "CustomerAgreementListLink",
    "DER",
    "DERAvailability",
    "DERAvailabilityLink",
    "DERCapability",
    "DERCapabilityLink",
    "DERControl",
    "DERControlBase",
    "DERControlList",
    "DERControlListLink",
    "DERControlResponse",
    "DERCurve",
    "DERCurveLink",
    "DERCurveList",
    "DERCurveListLink",
    "DERLink",
    "DERList",
    "DERListLink",
    "DERProgram",
    "DERProgramLink",
    "DERProgramList",
    "DERProgramListLink",
    "DERSettings",
    "DERSettingsLink",
    "DERStatus",
    "DERStatusLink",
    "DRLCCapabilities",
    "DateTimeInterval",
    "DefaultDERControl",
    "DefaultDERControlLink",
    "DemandResponseProgram",
    "DemandResponseProgramLink",
    "DemandResponseProgramList",
    "DemandResponseProgramListLink",
    "DeviceCapability",
    "DeviceCapabilityLink",
    "DeviceCategoryType",
    "DeviceInformation",
    "DeviceInformationLink",
    "DeviceStatus",
    "DeviceStatusLink",
    "DrResponse",
    "DutyCycle",
    "EndDevice",
    "EndDeviceControl",
    "EndDeviceControlList",
    "EndDeviceControlListLink",
    "EndDeviceLink",
    "EndDeviceList",
    "EndDeviceListLink",
    "EnvironmentalCost",
    "Error",
    "Event",
    "EventStatus",
    "File",
    "FileLink",
    "FileList",
    "FileListLink",
    "FileStatus",
    "FileStatusLink",
    "FixedPointType",
    "FixedVar",
    "FlowReservationRequest",
    "FlowReservationRequestList",
    "FlowReservationRequestListLink",
    "FlowReservationResponse",
    "FlowReservationResponseList",
    "FlowReservationResponseListLink",
    "FlowReservationResponseResponse",
    "FreqDroopType",
    "FunctionSetAssignments",
    "FunctionSetAssignmentsBase",
    "FunctionSetAssignmentsList",
    "FunctionSetAssignmentsListLink",
    "GPSLocationType",
    "HistoricalReading",
    "HistoricalReadingList",
    "HistoricalReadingListLink",
    "IEEE_802_15_4",
    "IPAddr",
    "IPAddrList",
    "IPAddrListLink",
    "IPInterface",
    "IPInterfaceList",
    "IPInterfaceListLink",
    "IdentifiedObject",
    "InverterStatusType",
    "LLInterface",
    "LLInterfaceList",
    "LLInterfaceListLink",
    "Link",
    "List_type",
    "ListLink",
    "LoadShedAvailability",
    "LoadShedAvailabilityList",
    "LoadShedAvailabilityListLink",
    "LocalControlModeStatusType",
    "LogEvent",
    "LogEventList",
    "LogEventListLink",
    "ManufacturerStatusType",
    "MessagingProgram",
    "MessagingProgramList",
    "MessagingProgramListLink",
    "MeterReading",
    "MeterReadingBase",
    "MeterReadingLink",
    "MeterReadingList",
    "MeterReadingListLink",
    "MirrorMeterReading",
    "MirrorMeterReadingList",
    "MirrorReadingSet",
    "MirrorUsagePoint",
    "MirrorUsagePointList",
    "MirrorUsagePointListLink",
    "Neighbor",
    "NeighborList",
    "NeighborListLink",
    "Notification",
    "NotificationList",
    "NotificationListLink",
    "Offset",
    "OperationalModeStatusType",
    "PEVInfo",
    "PowerConfiguration",
    "PowerFactor",
    "PowerFactorWithExcitation",
    "PowerStatus",
    "PowerStatusLink",
    "PrepayOperationStatus",
    "PrepayOperationStatusLink",
    "Prepayment",
    "PrepaymentLink",
    "PrepaymentList",
    "PrepaymentListLink",
    "PriceResponse",
    "PriceResponseCfg",
    "PriceResponseCfgList",
    "PriceResponseCfgListLink",
    "ProjectionReading",
    "ProjectionReadingList",
    "ProjectionReadingListLink",
    "RPLInstance",
    "RPLInstanceList",
    "RPLInstanceListLink",
    "RPLSourceRoutes",
    "RPLSourceRoutesList",
    "RPLSourceRoutesListLink",
    "RandomizableEvent",
    "RateComponent",
    "RateComponentLink",
    "RateComponentList",
    "RateComponentListLink",
    "ReactivePower",
    "ReactiveSusceptance",
    "Reading",
    "ReadingBase",
    "ReadingLink",
    "ReadingList",
    "ReadingListLink",
    "ReadingSet",
    "ReadingSetBase",
    "ReadingSetList",
    "ReadingSetListLink",
    "ReadingType",
    "ReadingTypeLink",
    "RealEnergy",
    "Registration",
    "RegistrationLink",
    "RequestStatus",
    "Resource",
    "RespondableIdentifiedObject",
    "RespondableResource",
    "RespondableSubscribableIdentifiedObject",
    "Response",
    "ResponseList",
    "ResponseListLink",
    "ResponseSet",
    "ResponseSetList",
    "ResponseSetListLink",
    "SelfDevice",
    "SelfDeviceLink",
    "ServiceChange",
    "ServiceSupplier",
    "ServiceSupplierLink",
    "ServiceSupplierList",
    "SetPoint",
    "SignedRealEnergy",
    "StateOfChargeStatusType",
    "StorageModeStatusType",
    "SubscribableIdentifiedObject",
    "SubscribableList",
    "SubscribableResource",
    "Subscription",
    "SubscriptionBase",
    "SubscriptionList",
    "SubscriptionListLink",
    "SupplyInterruptionOverride",
    "SupplyInterruptionOverrideList",
    "SupplyInterruptionOverrideListLink",
    "SupportedLocale",
    "SupportedLocaleList",
    "SupportedLocaleListLink",
    "TargetReading",
    "TargetReadingList",
    "TargetReadingListLink",
    "TargetReduction",
    "TariffProfile",
    "TariffProfileLink",
    "TariffProfileList",
    "TariffProfileListLink",
    "Temperature",
    "TextMessage",
    "TextMessageList",
    "TextMessageListLink",
    "TextResponse",
    "Time",
    "TimeConfiguration",
    "TimeLink",
    "TimeTariffInterval",
    "TimeTariffIntervalList",
    "TimeTariffIntervalListLink",
    "UnitValueType",
    "UnsignedFixedPointType",
    "UsagePoint",
    "UsagePointBase",
    "UsagePointLink",
    "UsagePointList",
    "UsagePointListLink",
    "VoltageRMS",
    "WattHour",
    "loWPAN",
]



import calendar
import time
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import IntEnum
from pathlib import Path
from typing import Union, Any, List, Dict

PathStr = Union[Path, str]
StrPath = PathStr
TimeType = int
TimeOffsetType = int
Lfdi = str

SEP_XML = "application/sep+xml"


def format_time(dt_obj: datetime, is_local: bool = False) -> TimeType:
    """ Return a proper IEEE2030_5 TimeType object for the dt_obj passed in.
            From IEEE 2030.5 spec:
                TimeType Object (Int64)
                    Time is a signed 64 bit value representing the number of seconds
                    since 0 hours, 0 minutes, 0 seconds, on the 1st of January, 1970,
                    in UTC, not counting leap seconds.
        :param dt_obj: Datetime object to convert to IEEE2030_5 TimeType object.
        :param is_local: dt_obj is in UTC or Local time. Default to UTC time.
        :return: Time XSD object
        :raises: If utc_dt_obj is not UTC
    """

    if dt_obj.tzinfo is None:
        raise Exception("IEEE 2030.5 times should be timezone aware UTC or local")

    if dt_obj.utcoffset() != timedelta(0) and not is_local:
        raise Exception("IEEE 2030.5 TimeType should be based on UTC")

    if is_local:
        return TimeType(int(time.mktime(dt_obj.timetuple())))
    else:
        return TimeType(int(calendar.timegm(dt_obj.timetuple())))


class DERControlType(IntEnum):
    # Control modes supported by the DER. Bit
    # positions SHALL be defined as follows:
    # 0 - Charge mode
    chargeMode = 0
    # 1 - Discharge mode
    dischargeMode = 1
    # 2 - opModConnect (Connect / Disconnect -
    # implies galvanic isolation)
    opModConnect = 2
    # 3 - opModEnergize (Energize / De-Energize)
    opModEnergize = 3
    # 4 - opModFixedPFAbsorbW (Fixed Power
    opModFixedPFAbsorb = 4
    # Factor Setpoint when absorbing active # power)
    # 5 - opModFixedPFInjectW (Fixed Power
    opModFixedPFInject = 5
    # Factor Setpoint when injecting active power)
    # 6 - opModFixedVar (Reactive Power Setpoint)
    opModFixedVar = 6
    # 7 - opModFixedW (Charge / Discharge Setpoint)
    opModFixedW = 7
    # 8 - opModFreqDroop (Frequency-Watt Parameterized Mode)
    opModFreqDroop = 8
    # 9 - opModFreqWatt (Frequency-Watt Curve Mode)
    opModFreqWatt = 9
    # 10 - opModHFRTMayTrip (High Frequency Ride Through, May Trip Mode)
    opModHFRTMayTrip = 10
    # 11 - opModHFRTMustTrip (High Frequency Ride Through, Must Trip Mode)
    opModHFRTMustTrip = 11
    # 12 - opModHVRTMayTrip (High Voltage Ride Through, May Trip Mode)
    opModHVRTMayTrip = 12
    # 13 - opModHVRTMomentaryCessation (High Voltage Ride Through, Momentary
    # Cessation Mode)
    opModHVRTMomentaryCessation = 13
    # 14 - opModHVRTMustTrip (High Voltage Ride Through, Must Trip Mode)
    opModHVRTMustTrip = 14
    # 15 - opModLFRTMayTrip (Low Frequency Ride Through)
    opModLFRTMayTrip = 15



from datetime import datetime, timedelta
from flask import Response
import pytz
import tzlocal

from ieee_2030_5.server.base_request import RequestOp
import ieee_2030_5.models as m
from ieee_2030_5.types_ import TimeOffsetType, format_time


class TimeRequest(RequestOp):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        # TODO fix for new stuff.
        # local_tz = datetime.now().astimezone().tzinfo
        # now_local = datetime.now().replace(tzinfo=local_tz)

        now_utc = datetime.utcnow().replace(tzinfo=pytz.utc)
        # now_utc = pytz.utc.localize(datetime.utcnow())
        local_tz = pytz.timezone(tzlocal.get_localzone().zone)
        now_local = datetime.now().replace(tzinfo=local_tz)

        start_dst_utc, end_dst_utc = [
            dt for dt in local_tz._utc_transition_times if dt.year == now_local.year
        ]

        utc_offset = local_tz.utcoffset(start_dst_utc - timedelta(days=1))
        dst_offset = local_tz.utcoffset(start_dst_utc + timedelta(days=1)) - utc_offset
        local_but_utc = datetime.now().replace(tzinfo=pytz.utc)

        tm = m.Time(currentTime=format_time(now_utc),
                    dstEndTime=format_time(end_dst_utc.replace(tzinfo=pytz.utc)),
                    dstOffset=TimeOffsetType(int(dst_offset.total_seconds())),
                    localTime=format_time(local_but_utc),
                    quality=None,
                    tzOffset=TimeOffsetType(utc_offset.total_seconds()))

        return self.build_response_from_dataclass(tm)